from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np

import dcmri.ui as ui
import dcmri.lib as lib
import dcmri.sig as sig
import dcmri.utils as utils
import dcmri.pk_aorta as pk_aorta
import dcmri.liver as liver


class AortaLiver(ui.Model):
    """Joint model for aorta and liver signals.

    This model uses a whole-body model to simultaneously predict signals in 
    aorta and liver.  

    For more detail on the whole-body model, see :ref:`whole-body-tissues`. 
    For more detail on the liver model, see :ref:`liver-tissues`. 

    Args:
        kinetics (str, optional): Tracer-kinetic liver model. See table 
          :ref:`table-liver-models` for options - only single-inlet models 
          are allowed. Defaults to '1I-IC-D'.
        stationary (str, optional): For intracellular tracers - stationarity 
          regime of the hepatocytes. The options are 'UE', 'E', 'U' or None. 
          For more detail see :ref:`liver-tissues`. Defaults to 'UE'.
        sequence (str, optional): imaging sequence. Possible values are 'SS'
          and 'SR'. Defaults to 'SS'.
        free (dict, optional): Dictionary with free parameters and their
          bounds. If not provided, a default set of free parameters is used.
          Defaults to None.
        params (dict, optional): values for the parameters of the tissue,
          specified as keyword parameters. Defaults are used for any that are
          not provided. See tables :ref:`AortaLiver-parameters` and
          :ref:`AortaLiver-defaults` for a list of parameters and their
          default values.

    See Also:
        `AortaLiver2scan`

    Example:

        Use the model to reconstruct concentrations from experimentally 
        derived signals.

    .. plot::
        :include-source:
        :context: close-figs

        >>> import matplotlib.pyplot as plt
        >>> import dcmri as dc

        Use `fake_tissue` to generate synthetic test data from 
        experimentally-derived concentrations:

        Use `fake_liver` to generate synthetic test data:

        >>> time, aif, vif, roi, gt = dc.fake_liver()

        Since this model generates two time curves, the x- and y-data are 
        tuples:

        >>> xdata, ydata = (time,time), (aif,roi)

        Build an aorta-liver model and parameters to match the 
        conditions of the fake liver data:

        >>> model = dc.AortaLiver(
        ...     dt = 0.5,
        ...     tmax = 180,
        ...     weight = 70,
        ...     agent = 'gadoxetate',
        ...     field_strength = 3.0,
        ...     dose = 0.2,
        ...     rate = 3,
        ...     t0 = 10,
        ...     TR = 0.005,
        ...     FA = 15,
        ... )

        Train the model on the data:

        >>> model.train(xdata, ydata, xtol=1e-3)

        Plot the reconstructed signals and concentrations and compare 
        against the experimentally derived data:

        >>> model.plot(xdata, ydata)

        We can also have a look at the model parameters after training:

        >>> model.print_params(round_to=3)
        -----------------------------------------
        Free parameters with their errors (stdev)
        -----------------------------------------
        Bolus arrival time (BAT): 17.127 (2.838) sec
        Cardiac output (CO): 211.047 (13.49) mL/sec
        Heart-lung mean transit time (Thl): 12.503 (3.527) sec
        Heart-lung transit time dispersion (Dhl): 0.465 (0.063)
        Organs mean transit time (To): 28.413 (10.294) sec
        Extraction fraction (Eb): 0.01 (0.623)
        Liver extracellular mean transit time (Te): 2.915 (0.588) sec
        Liver extracellular dispersion (De): 1.0 (0.19)
        Liver extracellular volume fraction (ve): 0.176 (0.015) mL/cm3
        Hepatocellular uptake rate (khe): 0.005 (0.001) mL/sec/cm3
        Hepatocellular transit time (Th): 600.0 (747.22) sec
        Organs extraction fraction (Eo): 0.261 (0.324)
        Organs extracellular mean transit time (Toe): 81.765 (329.097) sec
        ------------------
        Derived parameters
        ------------------
        Blood precontrast T1 (T10a): 1.629 sec
        Mean circulation time (Tc): 40.916 sec
        Liver precontrast T1 (T10l): 0.752 sec
        Biliary excretion rate (kbh): 0.001 mL/sec/cm3
        Hepatocellular tissue uptake rate (Khe): 0.026 mL/sec/cm3
        Biliary tissue excretion rate (Kbh): 0.002 mL/sec/cm3

    Notes:

        Table :ref:`AortaLiver-parameters` lists the parameters that are 
        relevant in each regime. Table :ref:`AortaLiver-defaults` list all 
        possible parameters and their default settings. 

        .. _AortaLiver-parameters:
        .. list-table:: **Aorta-Liver parameters**
            :widths: 20 30 30
            :header-rows: 1

            * - Parameters
              - When to use
              - Further detail
            * - dt, tmax
              - Always
              - Time axis for forward model
            * - t0, dose_tolerance
              - Always
              - For estimating baseline signal
            * - field_strength, weight, agent, dose, rate
              - Always
              - Injection protocol
            * - R10a, R10l, S0a, S0l 
              - Always
              - Precontrast R1 (:ref:`relaxation-params`) and 
                S0 (:ref:`params-per-sequence`)for aorta and liver 
            * - FA, TR, TS
              - Always
              - :ref:`params-per-sequence`
            * - TC
              - If **sequence** is 'SR'
              - :ref:`params-per-sequence`
            * - BAT, CO, Thl, Dhl, To, Eo, Tie, Eb
              - Always
              - :ref:`whole-body-tissues`
            * - H, ve, De
              - Always
              - :ref:`table-liver-models`
            * - khe, khe_i, kh_f, Th, Th_i, Th_f
              - Depends on **stationary**
              - :ref:`table-liver-models`

        .. _AortaLiver-defaults:
        .. list-table:: **Aorta-Liver parameter defaults**
            :widths: 5 10 10 10 10
            :header-rows: 1

            * - Parameter
              - Type
              - Value
              - Bounds
              - Free/Fixed
            * - 
              - **Simulation**
              -
              - 
              - 
            * - dt
              - Simulation
              - 0.5
              - [0, inf]
              - Fixed
            * - tmax
              - Simulation
              - 120
              - [0, inf]
              - Fixed
            * - dose_tolerance
              - Simulation
              - 0.1
              - [0, 1]
              - Fixed
            * - 
              - **Injection**
              -
              - 
              - 
            * - field_strength
              - Injection
              - 3
              - [0, inf]
              - Fixed
            * - weight
              - Injection
              - 70
              - [0, inf]
              - Fixed
            * - agent
              - Injection
              - 'gadoxetate'
              - None
              - Fixed
            * - dose
              - Injection
              - 0.0125
              - [0, inf]
              - Fixed
            * - rate
              - Injection
              - 1
              - [0, inf]
              - Fixed
            * - 
              - **Signal**
              -
              - 
              - 
            * - R10a
              - Signal
              - 0.7
              - [0, inf]
              - Fixed
            * - R10l
              - Signal
              - 0.7
              - [0, inf]
              - Fixed
            * - S0a
              - Signal
              - 1
              - [0, inf]
              - Free
            * - S0l
              - Signal
              - 1
              - [0, inf]
              - Free
            * - 
              - **Sequence**
              -
              - 
              - 
            * - FA
              - Sequence
              - 15
              - [0, inf]
              - Fixed
            * - S0
              - Sequence
              - 1
              - [0, inf]
              - Fixed
            * - TC
              - Sequence
              - 0.1
              - [0, inf]
              - Fixed
            * - TR
              - Sequence
              - 0.005
              - [0, inf]
              - Fixed
            * - TS
              - Sequence
              - 0
              - [0, inf]
              - Fixed
            * - 
              - **Whole body**
              -
              - 
              - 
            * - BAT
              - Whole body
              - 1200
              - [0, inf]
              - Free
            * - CO
              - Whole body
              - 100
              - [0, inf]
              - Free
            * - Thl
              - Whole body
              - 10
              - [0, 30]
              - Free
            * - Dhl
              - Whole body
              - 0.2
              - [0.05, 0.95]
              - Free
            * - To
              - Whole body
              - 20
              - [0, 60]
              - Free
            * - Eo
              - Whole body
              - 0.15
              - [0, 0.5]
              - Free
            * - Toe
              - Whole body
              - 120
              - [0, 800]
              - Free
            * - Eb
              - Whole body
              - 0.05
              - [0.01, 0.15]
              - Free
            * - 
              - **Liver**
              -
              - 
              - 
            * - H
              - Kinetic
              - 0.45
              - [0, 1]
              - Fixed
            * - Te
              - Kinetic
              - 30
              - [0.1, 60]
              - Free
            * - De
              - Kinetic
              - 0.85
              - [0, 1]
              - Free
            * - ve
              - Kinetic
              - 0.3
              - [0.01, 0.6]
              - Free
            * - khe
              - Kinetic
              - 0.003
              - [0, 0.1]
              - Free
            * - khe_i
              - Kinetic
              - 0.003
              - [0, 0.1]
              - Free
            * - khe_f
              - Kinetic
              - 0.003
              - [0, 0.1]
              - Free
            * - Th
              - Kinetic
              - 1800
              - [600, 36000]
              - Free
            * - Th_i
              - Kinetic
              - 1800
              - [600, 36000]
              - Free
            * - Th_f
              - Kinetic
              - 1800
              - [600, 36000]
              - Free
            * - vol
              - Kinetic
              - 1000
              - [0, 10000]
              - Free

    """

    def __init__(self, kinetics = '1I-IC-D', stationary='UE', sequence='SS', 
                 free=None, **params):

        # Configuration
        self.organs = '2cxm' # fixed
        self.kinetics = kinetics
        self.sequence = sequence 
        self.stationary = stationary

        self._check_config()
        self._set_defaults(free=free, **params)

        # Internal flags
        self._predict = None

    def _check_config(self):
        _check_config(self)

    def _params(self):
        return (PARAMS | PARAMS_WHOLE_BODY | PARAMS_SEQUENCE 
                | PARAMS_LIVER | PARAMS_DERIVED)
    
    def _model_pars(self):
        pars_sequence = {
            'SR': ['FA', 'TR', 'TS', 'TC'],
            'SS': ['FA', 'TR', 'TS'], 
        }
        pars = list(PARAMS.keys()) 
        pars += list(PARAMS_WHOLE_BODY.keys())
        pars += pars_sequence[self.sequence]
        pars += liver.params_liver(self.kinetics, self.stationary)
        pars += ['vol']
        return pars
    
    def _par_values(*args, **kwargs):
        return _par_values(*args, **kwargs)

    def _conc_aorta(self) -> np.ndarray:
        if self.organs == 'comp':
            organs = ['comp', (self.To,)]
        else:
            organs = ['2cxm', ([self.To, self.Toe], self.Eo)]
        self.t = np.arange(0, self.tmax, self.dt)
        conc = lib.ca_conc(self.agent)
        Ji = lib.ca_injection(
            self.t, self.weight, conc, self.dose, self.rate, self.BAT)
        Jb = pk_aorta.flux_aorta(
            Ji, E=self.Eb, dt=self.dt, tol=self.dose_tolerance,
            heartlung = ['pfcomp', (self.Thl, self.Dhl)],
            organs = organs)
        self.ca = Jb/self.CO
        return self.t, self.ca

    def _relax_aorta(self):
        return _relax_aorta(self)

    def _predict_aorta(self, xdata: np.ndarray) -> np.ndarray:
        self.tmax = max(xdata)+self.dt
        if self.TS is not None:
            self.tmax += self.TS
        t, R1b = self._relax_aorta()
        #seq = 'SRC' if self.sequence=='SR' else 'SS'
        pars = self._par_values(seq=self.sequence)
        signal = sig.signal(self.sequence, R1b, self.S0a, **pars)
        return utils.sample(xdata, t, signal, self.TS)

    def _conc_liver(*args, **kwargs):
        return _conc_liver(*args, **kwargs)

    def _relax_liver(*args, **kwargs):
        return _relax_liver(*args, **kwargs)

    def _predict_liver(self, xdata: np.ndarray) -> np.ndarray:
        t, R1l = self._relax_liver()
        pars = self._par_values(seq=self.sequence)
        signal = sig.signal(self.sequence, R1l, self.S0l, **pars)
        return utils.sample(xdata, t, signal, self.TS)

    def conc(self, sum=True):
        """Concentrations in aorta and liver.

        Args:
            sum (bool, optional): If set to true, the liver concentrations are 
              the sum over both compartments. If set to false, the 
              compartmental concentrations are returned individually. 
              Defaults to True.

        Returns:
            tuple: time points, aorta blood concentrations, liver 
              concentrations.
        """
        t, cb = self._conc_aorta()
        C = self._conc_liver(sum=sum)
        return t, cb, C

    def relax(self):
        """Relaxation rates in aorta and liver.

        Returns:
            tuple: time points, aorta blood R1, liver 
              R1.
        """
        t, R1b = self._relax_aorta()
        t, R1l = self._relax_liver()
        return t, R1b, R1l

    def predict(self, xdata: tuple) -> tuple:
        """Predict the data at given xdata

        Args:
            xdata (tuple): tuple of 2 arrays with time points for aorta and 
              liver, in that order. The two arrays can be different in length 
              and value.

        Returns:
            tuple: tuple of 2 arrays with signals for aorta and liver, in 
              that order. The arrays can be different in length and value but 
              each has to have the same length as its corresponding array of 
              time points.
        """
        # Public interface
        if self._predict is None:
            signala = self._predict_aorta(xdata[0])
            signall = self._predict_liver(xdata[1])
            return signala, signall
        # Private interface with different input & output types
        elif self._predict == 'aorta':
            return self._predict_aorta(xdata)
        elif self._predict == 'liver':
            return self._predict_liver(xdata)

    def train(self, xdata: tuple, ydata: tuple, joint=True, **kwargs):
        """Train the free parameters

        Args:
            xdata (tuple): tuple of 2 arrays with time points for aorta and 
              liver, in that order. The two arrays can be different in length 
              and value.
            ydata (array-like): tuple of 2 arrays with signals for aorta and 
              liver, in that order. The arrays can be different in length and 
              value but each has to have the same length as its corresponding 
              array of time points.
            joint (bool, optional): If True, the aorta and liver 
              parameters are trained simultaneously after training 
              them separately. Defaults to True.
            kwargs: any keyword parameters accepted by 
              `scipy.optimize.curve_fit`.

        Returns:
            Model: A reference to the model instance.
        """
        # Estimate BAT and S0a from data
        pars = self._par_values(seq=self.sequence)
        Srefb = sig.signal(self.sequence, self.R10a, 1, **pars)
        Srefl = sig.signal(self.sequence, self.R10l, 1, **pars)

        n0 = max([np.sum(xdata[0] < self.t0), 1])
        self.S0a = np.mean(ydata[0][:n0]) / Srefb
        self.S0l = np.mean(ydata[1][:n0]) / Srefl
        self.BAT = xdata[0][np.argmax(ydata[0])] - (1-self.Dhl)*self.Thl
        self.BAT = max([self.BAT, 0])

        # Copy the original free to restore later
        free = deepcopy(self.free)

        # Train free aorta parameters on aorta data
        self._predict = 'aorta'
        pars = list(PARAMS_WHOLE_BODY.keys())
        self.free = {s: free[s] for s in pars if s in free}
        ui.train(self, xdata[0], ydata[0], **kwargs)

        # Train free liver parameters on liver data
        self._predict = 'liver'
        pars = list(PARAMS_LIVER.keys())
        self.free = {s: free[s] for s in pars if s in free}
        ui.train(self, xdata[1], ydata[1], **kwargs)

        # Train all parameters on all data
        self._predict = None
        self.free = free
        if joint:
            return ui.train(self, xdata, ydata, **kwargs)
        else:
            return self

    def plot(self,
             xdata: tuple,
             ydata: tuple,
             xlim=None, ref=None,
             fname=None, show=True):
        """Plot the model fit against data

        Args:
            xdata (tuple): tuple of 2 arrays with time points for aorta and 
              liver, in that order. The two arrays can be different in length 
              and value.
            ydata (array-like): tuple of 2 arrays with signals for aorta and 
              liver, in that order. The arrays can be different in length and 
              value but each has to have the same length as its corresponding 
              array of time points.
            xlim (array_like, optional): 2-element array with lower and upper 
              boundaries of the x-axis. Defaults to None.
            ref (tuple, optional): Tuple of optional test data in the form 
              (x,y), where x is an array with x-values and y is an array with 
              y-values. Defaults to None.
            fname (path, optional): Filepath to save the image. If no value 
              is provided, the image is not saved. Defaults to None.
            show (bool, optional): If True, the plot is shown. Defaults to 
              True.
        """
        t, cb, C = self.conc(sum=False)
        sig = self.predict((t, t))
        fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, figsize=(10, 8))
        fig.subplots_adjust(wspace=0.3)
        _plot_data1scan(t, sig[0], xdata[0], ydata[0],
                        ax1, xlim,
                        color=['lightcoral', 'darkred'],
                        test=None if ref is None else ref[0])
        _plot_data1scan(t, sig[1], xdata[1], ydata[1],
                        ax3, xlim,
                        color=['cornflowerblue', 'darkblue'],
                        test=None if ref is None else ref[1])
        _plot_conc_aorta(t, cb, ax2, xlim)
        _plot_conc_liver(self, t, C, ax4, xlim)
        if fname is not None:
            plt.savefig(fname=fname)
        if show:
            plt.show()
        else:
            plt.close()

    def cost(self, xdata, ydata, metric='NRMS') -> float:
        """Return the goodness-of-fit

        Args:
            xdata (tuple): tuple of 2 arrays with time points for aorta and 
              liver, in that order. The two arrays can be different in length 
              and value.
            ydata (array-like): tuple of 2 arrays with signals for aorta and 
              liver, in that order. The arrays can be different in length and 
              value but each has to have the same length as its corresponding 
              array of time points.
            metric (str, optional): Which metric to use (see notes for 
              possible values). Defaults to 'NRMS'.

        Returns:
            float: goodness of fit.

        Notes:

            Available options are: 
            
            - 'RMS': Root-mean-square.
            - 'NRMS': Normalized root-mean-square. 
            - 'AIC': Akaike information criterion. 
            - 'cAIC': Corrected Akaike information criterion for small 
              models.
            - 'BIC': Baysian information criterion.
        """
        return super().cost(xdata, ydata, metric)


class AortaLiver2scan(ui.Model):
    """Joint model for aorta and liver signals measured over two scans.

    This model uses a whole-body model to simultaneously predict signals in 
    aorta and liver, measured over two separate scans.

    For more detail on the whole-body model, see :ref:`whole-body-tissues`. 
    For more detail on the liver model, see :ref:`liver-tissues`. 

    Args:
        kinetics (str, optional): Tracer-kinetic liver model. See table 
          :ref:`table-liver-models` for options - only single-inlet models 
          are allowed. Defaults to '1I-IC-D'.
        stationary (str, optional): For intracellular tracers - stationarity 
          regime of the hepatocytes. The options are 'UE', 'E', 'U' or None. 
          For more detail see :ref:`liver-tissues`. Defaults to 'UE'.
        stationary (str, optional): Stationarity regime of the hepatocytes. 
          The options are 'UE', 'E', 'U' or None. For more detail 
          see :ref:`liver-tissues`. Defaults to 'UE'.
        sequence (str, optional): imaging sequence. Possible values are 'SS'
          and 'SR'. Defaults to 'SS'.
        free (dict, optional): Dictionary with free parameters and their
          bounds. If not provided, a default set of free parameters is used.
          Defaults to None.
        params (dict, optional): values for the parameters of the tissue,
          specified as keyword parameters. Defaults are used for any that are
          not provided. See tables :ref:`AortaLiver2scan-parameters` and
          :ref:`AortaLiver2scan-defaults` for a list of parameters and their
          default values.

    See Also:
        `AortaLiver`

    Example:

        Use the model to reconstruct concentrations from experimentally 
        derived signals.

    .. plot::
        :include-source:
        :context: close-figs

        >>> import matplotlib.pyplot as plt
        >>> import dcmri as dc

        Use `fake_tissue` to generate synthetic test data from 
        experimentally-derived concentrations:

        >>> time, aif, roi, gt = dc.fake_tissue2scan(R10=1/dc.T1(3.0,'liver'))

        Since this model generates four time curves, the x- and y-data are 
        tuples:

        >>> xdata = (time[0], time[1], time[0], time[1])
        >>> ydata = (aif[0], aif[1], roi[0], roi[1])

        Build an aorta-liver model and parameters to match the conditions of 
        the fake tissue data:

        >>> model = dc.AortaLiver2scan(
        ...     dt = 0.5,
        ...     weight = 70,
        ...     agent = 'gadodiamide',
        ...     dose = 0.2,
        ...     dose2 = 0.2,
        ...     rate = 3,
        ...     field_strength = 3.0,
        ...     t0 = 10,
        ...     TR = 0.005,
        ...     FA = 15,
        ...     FA2 = 15,
        ... )

        Train the model on the data:

        >>> model.train(xdata, ydata, xtol=1e-3)

        Plot the reconstructed signals and concentrations and compare against 
        the experimentally derived data:

        >>> model.plot(xdata, ydata)

        We can also have a look at the model parameters after training:

        >>> model.print_params(round_to=3)
        -----------------------------------------
        Free parameters with their errors (stdev)  
        -----------------------------------------  
        Bolus arrival time (BAT): 17.13 (1.771) sec
        Cardiac output (CO): 208.547 (9.409) mL/sec
        Heart-lung mean transit time (Thl): 12.406 (2.137) sec
        Heart-lung transit time dispersion (Dhl): 0.459 (0.04)
        Organs mean transit time (To): 30.912 (3.999) sec
        Extraction fraction (Eb): 0.064 (0.032)
        Liver extracellular mean transit time (Te): 2.957 (0.452) sec
        Liver extracellular dispersion (De): 1.0 (0.146)
        Liver extracellular volume fraction (ve): 0.077 (0.007) mL/cm3
        Hepatocellular uptake rate (khe): 0.002 (0.001) mL/sec/cm3
        Hepatocellular transit time (Th): 600.0 (1173.571) sec
        Organs extraction fraction (Eo): 0.2 (0.057)
        Organs extracellular mean transit time (Toe): 87.077 (56.882) sec
        Hepatocellular uptake rate (final) (khe_f): 0.001 (0.001) mL/sec/cm3
        Hepatocellular transit time (final) (Th_f): 600.0 (623.364) sec
        ------------------
        Derived parameters
        ------------------
        Blood precontrast T1 (T10a): 1.629 sec
        Mean circulation time (Tc): 43.318 sec
        Liver precontrast T1 (T10l): 0.752 sec
        Biliary excretion rate (kbh): 0.002 mL/sec/cm3
        Hepatocellular tissue uptake rate (Khe): 0.023 mL/sec/cm3
        Biliary tissue excretion rate (Kbh): 0.002 mL/sec/cm3
        Hepatocellular uptake rate (initial) (khe_i): 0.003 mL/sec/cm3
        Hepatocellular transit time (initial) (Th_i): 600.0 sec
        Hepatocellular uptake rate variance (khe_var): 0.934
        Biliary tissue excretion rate variance (Kbh_var): 0.0
        Biliary excretion rate (initial) (kbh_i): 0.002 mL/sec/cm3
        Biliary excretion rate (final) (kbh_f): 0.002 mL/sec/cm3

    Notes:

        Table :ref:`AortaLiver-parameters` lists the parameters that are 
        relevant in each regime. Table :ref:`AortaLiver-defaults` list all 
        possible parameters and their default settings. 

        .. _AortaLiver2scan-parameters:
        .. list-table:: **Aorta-Liver 2-scan parameters**
            :widths: 20 30 30
            :header-rows: 1

            * - Parameters
              - When to use
              - Further detail
            * - dt, tmax
              - Always
              - Time axis for forward model
            * - t0, dose_tolerance
              - Always
              - For estimating baseline signal
            * - field_strength, weight, agent, dose, dose2, rate
              - Always
              - Injection protocol
            * - R10a, R102a, R10l, R102l, S0a, S02a, S0l, S02l
              - Always
              - Precontrast R1 (:ref:`relaxation-params`) and 
                S0 (:ref:`params-per-sequence`) for aorta and liver 
            * - FA, TR, TS, FA2
              - Always
              - :ref:`params-per-sequence`
            * - TC
              - If **sequence** is 'SR'
              - :ref:`params-per-sequence`
            * - BAT, BAT2, CO, Thl, Dhl, To, Eo, Tie, Eb
              - Always
              - :ref:`whole-body-tissues`
            * - H, ve, De
              - Always
              - :ref:`table-liver-models`
            * - khe, khe_i, kh_f, Th, Th_i, Th_f
              - Depends on **stationary**
              - :ref:`table-liver-models`

        .. _AortaLiver2scan-defaults:
        .. list-table:: **Aorta-Liver 2-scan parameter defaults**
            :widths: 5 10 10 10 10
            :header-rows: 1

            * - Parameter
              - Type
              - Value
              - Bounds
              - Free/Fixed
            * - 
              - **Simulation**
              -
              - 
              - 
            * - dt
              - Simulation
              - 0.5
              - [0, inf]
              - Fixed
            * - tmax
              - Simulation
              - 120
              - [0, inf]
              - Fixed
            * - dose_tolerance
              - Simulation
              - 0.1
              - [0, 1]
              - Fixed
            * - 
              - **Injection**
              -
              - 
              - 
            * - field_strength
              - Injection
              - 3
              - [0, inf]
              - Fixed
            * - weight
              - Injection
              - 70
              - [0, inf]
              - Fixed
            * - agent
              - Injection
              - 'gadoxetate'
              - None
              - Fixed
            * - dose
              - Injection
              - 0.0125
              - [0, inf]
              - Fixed
            * - rate
              - Injection
              - 1
              - [0, inf]
              - Fixed
            * - 
              - **Signal**
              -
              - 
              - 
            * - R10a
              - Signal
              - 0.7
              - [0, inf]
              - Fixed
            * - R10l
              - Signal
              - 0.7
              - [0, inf]
              - Fixed
            * - S0a
              - Signal
              - 1
              - [0, inf]
              - Free
            * - S0l
              - Signal
              - 1
              - [0, inf]
              - Free
            * - 
              - **Sequence**
              -
              - 
              - 
            * - FA
              - Sequence
              - 15
              - [0, inf]
              - Fixed
            * - FA2
              - Sequence
              - 15
              - [0, inf]
              - Fixed
            * - S0
              - Sequence
              - 1
              - [0, inf]
              - Fixed
            * - TC
              - Sequence
              - 0.1
              - [0, inf]
              - Fixed
            * - TR
              - Sequence
              - 0.005
              - [0, inf]
              - Fixed
            * - TS
              - Sequence
              - 0
              - [0, inf]
              - Fixed
            * - 
              - **Whole body**
              -
              - 
              - 
            * - BAT
              - Whole body
              - 1200
              - [0, inf]
              - Free
            * - CO
              - Whole body
              - 100
              - [0, inf]
              - Free
            * - Thl
              - Whole body
              - 10
              - [0, 30]
              - Free
            * - Dhl
              - Whole body
              - 0.2
              - [0.05, 0.95]
              - Free
            * - To
              - Whole body
              - 20
              - [0, 60]
              - Free
            * - Eo
              - Whole body
              - 0.15
              - [0, 0.5]
              - Free
            * - Toe
              - Whole body
              - 120
              - [0, 800]
              - Free
            * - Eb
              - Whole body
              - 0.05
              - [0.01, 0.15]
              - Free
            * - 
              - **Liver**
              -
              - 
              - 
            * - H
              - Kinetic
              - 0.45
              - [0, 1]
              - Fixed
            * - Te
              - Kinetic
              - 30
              - [0.1, 60]
              - Free
            * - De
              - Kinetic
              - 0.85
              - [0, 1]
              - Free
            * - ve
              - Kinetic
              - 0.3
              - [0.01, 0.6]
              - Free
            * - khe
              - Kinetic
              - 0.003
              - [0, 0.1]
              - Free
            * - khe_i
              - Kinetic
              - 0.003
              - [0, 0.1]
              - Free
            * - khe_f
              - Kinetic
              - 0.003
              - [0, 0.1]
              - Free
            * - Th
              - Kinetic
              - 1800
              - [600, 36000]
              - Free
            * - Th_i
              - Kinetic
              - 1800
              - [600, 36000]
              - Free
            * - Th_f
              - Kinetic
              - 1800
              - [600, 36000]
              - Free
            * - vol
              - Kinetic
              - 1000
              - [0, 10000]
              - Free
    """

    def __init__(self, kinetics = '1I-IC-D', stationary=None, sequence='SS', 
                 free=None, **params):

        # Configuration
        self.organs = '2cxm' # fixed
        self.kinetics = kinetics
        self.sequence = sequence 
        self.stationary = stationary

        self._check_config()
        self._set_defaults(free=free, **params)

        # Internal flags
        self._predict = None

    def _check_config(self):
        _check_config(self)
        
    def _params(self):
        return (PARAMS | PARAMS_2SCAN | PARAMS_WHOLE_BODY | PARAMS_SEQUENCE 
                | PARAMS_LIVER | PARAMS_DERIVED)

    def _model_pars(self):
        pars_sequence = {
            'SR': ['FA', 'TR', 'TS', 'TC'],
            'SS': ['FA', 'TR', 'TS'], 
        }
        pars = list(PARAMS.keys()) 
        pars += list(PARAMS_2SCAN.keys())
        pars += list(PARAMS_WHOLE_BODY.keys())
        pars += pars_sequence[self.sequence]
        pars += liver.params_liver(self.kinetics, self.stationary)
        pars += ['vol']
        return pars
    
    def _sequence_parameters_first_scan(self):
        if self.sequence == 'SR':
            return {
                'FA': self.FA, 
                'TR': self.TR, 
                'TC': self.TC, 
            }
        if self.sequence == 'SS':
            return {
                'FA': self.FA, 
                'TR': self.TR, 
            }
        if self.sequence == 'SRC':
            return {
                'TC': self.TC, 
            }

    def _sequence_parameters_second_scan(self):
        if self.sequence == 'SR': 
            return {
                'FA': self.FA2, 
                'TR': self.TR, 
                'TC': self.TC, 
            }
        if self.sequence == 'SS':
            return {
                'FA': self.FA2, 
                'TR': self.TR, 
            }
        if self.sequence == 'SRC':
            return {
                'TC': self.TC, 
            }
    
    def _par_values(*args, **kwargs):
        return _par_values(*args, **kwargs)

    def _conc_aorta(self) -> tuple[np.ndarray, np.ndarray]:
        if self.organs == 'comp':
            organs = ['comp', (self.To,)]
        else:
            organs = ['2cxm', ([self.To, self.Toe], self.Eo)]
        self.t = np.arange(0, self.tmax, self.dt)
        conc = lib.ca_conc(self.agent)
        J1 = lib.ca_injection(
            self.t, self.weight, conc, self.dose, self.rate, self.BAT)
        J2 = lib.ca_injection(
            self.t, self.weight, conc, self.dose2, self.rate, self.BAT2)
        Jb = pk_aorta.flux_aorta(
            J1 + J2, E=self.Eb, dt=self.dt, tol=self.dose_tolerance,
            heartlung = ['pfcomp', (self.Thl, self.Dhl)],
            organs = organs)
        self.ca = Jb/self.CO
        return self.t, self.ca
    
    def _relax_aorta(self):
        return _relax_aorta(self)

    def _predict_aorta(self,
                       xdata: tuple[np.ndarray, np.ndarray],
                       ) -> tuple[np.ndarray, np.ndarray]:
        self.tmax = max(xdata[1])+self.dt
        if self.TS is not None:
            self.tmax += self.TS
        t, R1 = self._relax_aorta()
        t1 = t <= xdata[0][-1]
        t2 = t >= xdata[1][0]
        R11 = R1[t1]
        R12 = R1[t2]
        #seq = 'SRC' if self.sequence=='SR' else 'SS'
        #pars = self._par_values(seq=self.sequence)
        pars = self._sequence_parameters_first_scan()
        signal1 = sig.signal(self.sequence, R11, self.S0a, **pars)
        pars = self._sequence_parameters_second_scan()
        signal2 = sig.signal(self.sequence, R12, self.S02a, **pars)
        return (
            utils.sample(xdata[0], t[t1], signal1, self.TS),
            utils.sample(xdata[1], t[t2], signal2, self.TS),
        )

    def _conc_liver(*args, **kwargs):
        return _conc_liver(*args, **kwargs)

    def _relax_liver(*args, **kwargs):
        return _relax_liver(*args, **kwargs)

    def _predict_liver(self, xdata: tuple[np.ndarray, np.ndarray],
                       ) -> tuple[np.ndarray, np.ndarray]:
        t, R1 = self._relax_liver()
        t1 = t <= xdata[0][-1]
        t2 = t >= xdata[1][0]
        R11 = R1[t1]
        R12 = R1[t2]
        #pars = self._par_values(seq=self.sequence)
        pars = self._sequence_parameters_first_scan()
        signal1 = sig.signal(self.sequence, R11, self.S0l, **pars)
        pars = self._sequence_parameters_second_scan()
        signal2 = sig.signal(self.sequence, R12, self.S02l, **pars)
        return (
            utils.sample(xdata[0], t[t1], signal1, self.TS),
            utils.sample(xdata[1], t[t2], signal2, self.TS),
        )
    
    def conc(self, sum=True):
        """Concentrations in aorta and liver.

        Args:
            sum (bool, optional): If set to true, the liver concentrations 
              are the sum over both compartments. If set to false, the 
              compartmental concentrations are returned individually. 
              Defaults to True.

        Returns:
            tuple: time points, aorta blood concentrations, liver 
              concentrations.
        """
        t, cb = self._conc_aorta()
        C = self._conc_liver(sum=sum)
        return t, cb, C
    
    def relax(self):
        """Relaxation rates in aorta and liver.

        Returns:
            tuple: time points, aorta blood R1, liver R1.
        """
        t, R1b = self._relax_aorta()
        t, R1l = self._relax_liver()
        return t, R1b, R1l

    def predict(self, xdata: tuple) -> tuple:
        """Predict the data at given time points

        Args:
            xdata (tuple): tuple of 4 arrays with time points for aorta in 
              the first scan, aorta in the second stand, liver in the first 
              scan, and liver in the second scan, in that order. The four 
              arrays can be different in length and value.

        Returns:
            tuple: tuple of 4 arrays with signals for aorta in the first 
            scan, aorta in the second stand, liver in the first scan, and 
            liver in the second scan, in that order. The arrays have the 
            same length as its corresponding array of time points.
        """
        # Public interface
        if self._predict is None:
            signal_a = self._predict_aorta((xdata[0], xdata[1]))
            signal_l = self._predict_liver((xdata[2], xdata[3]))
            return signal_a + signal_l
        # Private interface with different in- and outputs
        elif self._predict == 'aorta':
            return self._predict_aorta(xdata)
        elif self._predict == 'liver':
            return self._predict_liver(xdata)

    def train(self, xdata: tuple, ydata: tuple, joint=True, **kwargs):
        # x,y: (aorta scan 1, aorta scan 2, liver scan 1, liver scan 2)
        """Train the free parameters

        Args:
            xdata (tuple): tuple of 4 arrays with time points for aorta in 
              the first scan, aorta in the second stand, liver in the first 
              scan, and liver in the second scan, in that order. The four 
              arrays can be different in length and value.
            ydata (tuple): tuple of 4 arrays with signals for aorta in the 
              first scan, aorta in the second stand, liver in the first scan, 
              and liver in the second scan, in that order. The arrays can be 
              different in length but each has to have the same length as its 
              corresponding array of time points.
            joint (bool, optional): If True, aorta and liver parameters 
              are trained jointly after training them separately. 
              Defaults to True.
            kwargs: any keyword parameters accepted by 
              `scipy.optimize.curve_fit`.

        Returns:
            AortaLiver2scan: A reference to the model instance.
        """
        # Estimate BAT
        T, D = self.Thl, self.Dhl
        self.BAT = xdata[0][np.argmax(ydata[0])] - (1-D)*T
        self.BAT2 = xdata[1][np.argmax(ydata[1])] - (1-D)*T

        # Estimate S0
        pars = self._sequence_parameters_first_scan()
        Srefb = sig.signal(self.sequence, self.R10a, 1, **pars)
        Srefl = sig.signal(self.sequence, self.R10l, 1, **pars)
        pars = self._sequence_parameters_second_scan()
        Sref2l = sig.signal(self.sequence, self.R102l, 1, **pars)
        Sref2b = sig.signal(self.sequence, self.R102a, 1, **pars)

        n0 = max([np.sum(xdata[0] < self.t0), 2])
        self.S0a = np.mean(ydata[0][1:n0]) / Srefb
        self.S02a = np.mean(ydata[1][1:n0]) / Sref2b
        self.S0l = np.mean(ydata[2][1:n0]) / Srefl
        self.S02l = np.mean(ydata[3][1:n0]) / Sref2l

        free = deepcopy(self.free)

        # Train free aorta parameters on aorta data
        self._predict = 'aorta'
        pars = list(PARAMS_WHOLE_BODY.keys()) + ['BAT2', 'S02a']
        self.free = {s: free[s] for s in pars if s in free}
        ui.train(self, (xdata[0], xdata[1]), (ydata[0], ydata[1]), **kwargs)

        # Train free liver parameters on liver data
        self._predict = 'liver'
        pars = list(PARAMS_LIVER.keys()) + ['S02l']
        self.free = {s: free[s] for s in pars if s in free}
        ui.train(self, (xdata[2], xdata[3]), (ydata[2], ydata[3]), **kwargs)

        # Train all parameters on all data
        self._predict = None
        self.free = free
        if joint:
            return ui.train(self, xdata, ydata, **kwargs)
        else:
            return self
    

    def plot(self, xdata: tuple, ydata: tuple,
             ref=None, xlim=None, fname=None, show=True):
        """Plot the model fit against data

        Args:
            xdata (tuple): tuple of 4 arrays with time points for aorta in the 
              first scan, aorta in the second stand, liver in the first scan, 
              and liver in the second scan, in that order. The four arrays can 
              be different in length and value.
            ydata (tuple): tuple of 4 arrays with signals for aorta in the 
              first scan, aorta in the second stand, liver in the first scan, 
              and liver in the second scan, in that order. The arrays can be 
              different in length but each has to have the same length as its 
              corresponding array of time points.
            xlim (array_like, optional): 2-element array with lower and upper 
              boundaries of the x-axis. Defaults to None.
            ref (tuple, optional): Tuple of optional test data in the form 
              (x,y), where x is an array with x-values and y is an array with 
              y-values. Defaults to None.
            fname (path, optional): Filepath to save the image. If no value 
              is provided, the image is not saved. Defaults to None.
            show (bool, optional): If True, the plot is shown. Defaults to 
              True.
        """

        t, cb, C = self.conc(sum=False)
        ta1 = t[t <= xdata[1][0]]
        ta2 = t[(t > xdata[1][0]) & (t <= xdata[1][-1])]
        tl1 = t[t <= xdata[3][0]]
        tl2 = t[(t > xdata[3][0]) & (t <= xdata[3][-1])]
        sig = self.predict((ta1, ta2, tl1, tl2))
        fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, figsize=(10, 8))
        fig.subplots_adjust(wspace=0.3)
        _plot_data2scan((ta1, ta2), sig[:2], xdata[:2], ydata[:2],
                        ax1, xlim,
                        color=['lightcoral', 'darkred'],
                        test=None if ref is None else ref[0])
        _plot_data2scan((tl1, tl2), sig[2:], xdata[2:], ydata[2:],
                        ax3, xlim,
                        color=['cornflowerblue', 'darkblue'],
                        test=None if ref is None else ref[1])
        _plot_conc_aorta(t, cb, ax2, xlim)
        _plot_conc_liver(self, t, C, ax4, xlim)
        if fname is not None:
            plt.savefig(fname=fname)
        if show:
            plt.show()
        else:
            plt.close()

    def cost(self, xdata: tuple, ydata: tuple, metric='NRMS') -> float:
        """Return the goodness-of-fit

        Args:
            xdata (tuple): tuple of 4 arrays with time points for aorta in the 
              first scan, aorta in the second stand, liver in the first scan, 
              and liver in the second scan, in that order. The four arrays can 
              be different in length and value.
            ydata (tuple): tuple of 4 arrays with signals for aorta in the 
              first scan, aorta in the second stand, liver in the first scan, 
              and liver in the second scan, in that order. The arrays can be 
              different in length but each has to have the same length as its 
              corresponding array of time points.
            metric (str, optional): Which metric to use (see notes for 
              possible values). Defaults to 'NRMS'.

        Returns:
            float: goodness of fit.

        Notes:

            Available options are: 
            
            - 'RMS': Root-mean-square.
            - 'NRMS': Normalized root-mean-square. 
            - 'AIC': Akaike information criterion. 
            - 'cAIC': Corrected Akaike information criterion for small 
              models.
            - 'BIC': Baysian information criterion.
        """
        return super().cost(xdata, ydata, metric)


def _relax_aorta(self) -> np.ndarray:
    t, cb = self._conc_aorta()
    rb = lib.relaxivity(self.field_strength, 'blood', self.agent)
    return t, self.R10a + rb*cb

def _conc_liver(self, sum=True):
    pars = self._par_values(kin=True)
    return liver.conc_liver(self.ca, dt=self.dt, sum=sum, **pars)
    
def _relax_liver(self):
    t = np.arange(0, self.tmax, self.dt)
    Cl = self._conc_liver(sum=False)
    rp = lib.relaxivity(self.field_strength, 'plasma', self.agent)
    rh = lib.relaxivity(self.field_strength, 'hepatocytes', self.agent)
    if 'IC' in self.kinetics:
        return t, self.R10l + rp*Cl[0, :] + rh*Cl[1, :]
    else:
        return t, self.R10l + rp*Cl




# Helper functions for plotting

def _plot_conc_aorta(t: np.ndarray, cb: np.ndarray, ax, xlim=None):
    if xlim is None:
        xlim = [t[0], t[-1]]
    ax.set(xlabel='Time (min)', ylabel='Concentration (mM)',
           xlim=np.array(xlim)/60)
    ax.plot(t/60, 0*t, color='gray')
    ax.plot(t/60, 1000*cb, linestyle='-',
            color='darkred', linewidth=2.0, label='Aorta')
    ax.legend()

def _plot_conc_liver(self, t, C, ax, xlim=None):
    color = 'darkblue'
    if xlim is None:
        xlim = [t[0], t[-1]]
    ax.set(xlabel='Time (min)', ylabel='Tissue concentration (mM)',
           xlim=np.array(xlim)/60)
    ax.plot(t/60, 0*t, color='gray')
    if 'IC' in self.kinetics:
        ax.plot(t/60, 1000*C[0, :], linestyle='-.',
                color=color, linewidth=2.0, label='Extracellular')
        ax.plot(t/60, 1000*C[1, :], linestyle='--',
                color=color, linewidth=2.0, label='Hepatocytes')
        ax.plot(t/60, 1000*(C[0, :]+C[1, :]), linestyle='-',
                color=color, linewidth=2.0, label='Tissue')
    else:
        ax.plot(t/60, 1000*C, linestyle='-',
                color=color, linewidth=2.0, label='Tissue')        
    ax.legend()

def _plot_data2scan(t: tuple[np.ndarray, np.ndarray], 
                    sig: tuple[np.ndarray, np.ndarray],
                    xdata: tuple[np.ndarray, np.ndarray], 
                    ydata: tuple[np.ndarray, np.ndarray],
                    ax, xlim, color=['black', 'black'], test=None):
    if xlim is None:
        xlim = [0, t[1][-1]]
    ax.set(xlabel='Time (min)', ylabel='MR Signal (a.u.)', 
           xlim=np.array(xlim)/60)
    ax.plot(np.concatenate(xdata)/60, np.concatenate(ydata),
            marker='o', color=color[0], label='fitted data', linestyle='None')
    ax.plot(np.concatenate(t)/60, np.concatenate(sig),
            linestyle='-', color=color[1], linewidth=3.0, label='fit')
    if test is not None:
        ax.plot(np.array(test[0])/60, test[1], color='black',
                marker='D', linestyle='None', label='Test data')
    ax.legend()


def _plot_data1scan(t: np.ndarray, sig: np.ndarray,
                    xdata: np.ndarray, ydata: np.ndarray,
                    ax, xlim, color=['black', 'black'],
                    test=None):
    if xlim is None:
        xlim = [t[0], t[-1]]
    ax.set(xlabel='Time (min)', ylabel='MR Signal (a.u.)', 
           xlim=np.array(xlim)/60)
    ax.plot(xdata/60, ydata, marker='o',
            color=color[0], label='fitted data', linestyle='None')
    ax.plot(t/60, sig, linestyle='-',
            color=color[1], linewidth=3.0, label='fit')
    if test is not None:
        ax.plot(np.array(test[0])/60, test[1], color='black',
                marker='D', linestyle='None', label='Test data')
    ax.legend()



def _check_config(self):
    if self.sequence not in ['SS', 'SR']:
        raise ValueError(
            'Sequence ' + str(self.sequence) + ' is not available.')
    if self.kinetics[0] != '1':
        raise ValueError('Only single-inlet models are allowed.')
    liver.params_liver(self.kinetics, self.stationary)


def _par_values(self, kin=False, export=False, seq=None):

    if seq is not None:
        pars = {
            'SR': ['FA', 'TR', 'TC'],
            'SS': ['FA', 'TR'], 
            'SRC': ['TC'],
        }
        pars = pars[self.sequence]
        return {par: getattr(self, par) for par in pars}
    
    if kin:
        pars = liver.params_liver(self.kinetics, self.stationary)
        return {par: getattr(self, par) for par in pars}
    
    if export:
        pars = self._par_values()
        all = self._model_pars()
        p1 = liver.params_liver(self.kinetics, self.stationary)
        p2 = list(PARAMS_WHOLE_BODY.keys())
        p3 = list(PARAMS_DERIVED.keys())
        retain = p1 + p2 + p3 + ['S02a', 'S02l', 'BAT2']
        discard = set(all) - set(retain)
        return {p: pars[p] for p in pars if p not in discard}
    
    pars = self._model_pars()
    p = {par: getattr(self, par) for par in pars}

    try:
        p['Th'] = np.mean([p['Th_i'], p['Th_f']])
    except KeyError:
        pass
    try:
        p['khe'] = np.mean([p['khe_i'], p['khe_f']])
    except KeyError:
        pass
    try:
        p['Kbh'] = _div(1, p['Th'])
    except KeyError:
        pass
    try:
        p['Khe'] = _div(p['khe'], p['ve'])
    except KeyError:
        pass
    try:
        p['kbh'] = _div(1-p['ve'], p['Th'])
    except KeyError:
        pass
    try:
        p['kbh_i'] = _div(1-p['ve'], p['Th_i'])
    except KeyError:
        pass
    try:
        p['kbh_f'] = _div(1-p['ve'], p['Th_f'])
    except KeyError:
        pass
        try:
            p['E'] = p['khe']/(p['khe']+p['Fp'])
        except KeyError:
            try:
                p['E'] = p['khe']/(p['khe']+self.Fp)
            except:
                pass
        try:
            p['Ktrans'] = (1-p['E'])*p['khe']
        except KeyError:
            pass   
    if p['vol'] is not None:
        try:
            p['CL'] = p['khe']*p['vol']
        except KeyError:
            pass
        
    return p


PARAMS = {

    # Prediction and training
    't0': {
        'init': 0,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Baseline duration',
        'unit': 'sec',
    },
    'dt': {
        'init': 0.5,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Forward model time step',
        'unit': 'sec',
    },
    'tmax': {
        'init': 120,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Maximum acquisition time',
        'unit': 'sec',
    },
    'dose_tolerance': {
        'init': 0.1,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Dose tolerance',
        'unit': '',
    },


    # Injection
    'field_strength': {
        'init': 3.0,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Magnetic field strength',
        'unit': 'T',
    },
    'weight': {
        'init': 70.0,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Subject weight',
        'unit': 'kg',
    },
    'agent': {
        'init': 'gadoxetate',
        'default_free': False,
        'bounds': None,
        'name': 'Contrast agent',
        'unit': None,
    },
    'dose': {
        'init': lib.ca_std_dose('gadoxetate')/2,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'First contrast agent dose',
        'unit': 'mL/kg',
    },
    'rate': {
        'init': 1,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Contrast agent injection rate',
        'unit': 'mL/sec',
    },


    # Signal
    'R10a': {
        'init': 1/lib.T1(3.0, 'blood'),
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Aorta first baseline R1',
        'unit': 'Hz',
    },
    'S0a': {
        'init': 1,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Aorta first signal scale factor',
        'unit': 'a.u.',
    },
    'R10l': {
        'init': 1/lib.T1(3.0, 'liver'),
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Liver first baseline R1',
        'unit': 'Hz',
    },
    'S0l': {
        'init': 1,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Liver first signal scale factor',
        'unit': 'a.u.',
    },

}

PARAMS_2SCAN = {
    'dose2': {
        'init': lib.ca_std_dose('gadoxetate')/2,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Second contrast agent dose',
        'unit': 'mL/kg',
    },
    'FA2': {
        'init': 15.0,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Second flip angle',
        'unit': 'deg',
    },
    'R102a': {
        'init': 1/lib.T1(3.0, 'blood'),
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Aorta second baseline R1',
        'unit': 'Hz',
    },
    'S02a': {
        'init': 1,
        'default_free': True,
        'bounds': [0, np.inf],
        'name': 'Aorta second signal scale factor',
        'unit': 'a.u.',
    },
    'R102l': {
        'init': 1/lib.T1(3.0, 'liver'),
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Liver second baseline R1',
        'unit': 'Hz',
    },
    'S02l': {
        'init': 1,
        'default_free': True,
        'bounds': [0, np.inf],
        'name': 'Liver second signal scale factor',
        'unit': 'a.u.',
    },
    'BAT2': {
        'init': 1200,
        'default_free': True,
        'bounds': [0, np.inf],
        'name': 'Second bolus arrival time',
        'unit': 'sec',
    },
}

PARAMS_WHOLE_BODY = {

    # Body
    'BAT': {
        'init': 60,
        'default_free': True,
        'bounds': [0, np.inf],
        'name': 'First bolus arrival time',
        'unit': 'sec',
    },
    'CO': {
        'init': 100,
        'default_free': True,
        'bounds': [0, 300],
        'name': 'Cardiac output',
        'unit': 'mL/sec',
    },
    'Thl': {
        'init': 10,
        'default_free': True,
        'bounds': [0, 30],
        'name': 'Heart-lung mean transit time',
        'unit': 'sec',
    },
    'Dhl': {
        'init': 0.2,
        'default_free': True,
        'bounds': [0.05, 0.95],
        'name': 'Heart-lung dispersion',
        'unit': '',
    },
    'To': {
        'init': 20,
        'default_free': True,
        'bounds': [0, 60],
        'name': 'Organs blood mean transit time',
        'unit': 'sec',
    },
    'Eo': {
        'init': 0.15,
        'default_free': True,
        'bounds': [0, 0.5],
        'name': 'Organs extraction fraction',
        'unit': '',
    },
    'Toe': {
        'init': 120,
        'default_free': True,
        'bounds': [0, 800],
        'name': 'Organs extravascular mean transit time',
        'unit': 'sec',
    },
    'Eb': {
        'init': 0.05,
        'default_free': True,
        'bounds': [0.01, 0.15],
        'name': 'Body extraction fraction',
        'unit': '',
    },
}


PARAMS_SEQUENCE = {
    'TR': {
        'init': 0.005,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Repetition time',
        'unit': 'sec',
    },
    'FA': {
        'init': 15.0,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Flip angle',
        'unit': 'deg',
    },
    'TC': {
        'init': 0.180,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Time to center',
        'unit': 'sec',
    },
    'TS': {
        'init': None,
        'default_free': False,
        'bounds': [0, np.inf],
        'name': 'Sampling time',
        'unit': 'sec',
    },
}

PARAMS_LIVER = {
    'H': {
        'init': 0.45,
        'default_free': False,
        'bounds': [0, 1],
        'name': 'Hematocrit',
        'unit': '',
    },
    'Te': {
        'init': 30.0,
        'default_free': True,
        'bounds': [0.1, 60],
        'name': 'Extracellular mean transit time',
        'unit': 'sec',
    },
    'De': {
        'init': 0.85,
        'default_free': True,
        'bounds': [0, 1],
        'name': 'Extracellular dispersion',
        'unit': '',
    },
    've': {
        'init': 0.3,
        'default_free': True,
        'bounds': [0.01, 0.6],
        'name': 'Liver extracellular volume fraction',
        'unit': 'mL/cm3',
    },
    'khe': {
        'init': 0.003,
        'default_free': True,
        'bounds': [0.0, 0.1],
        'name': 'Hepatocellular uptake rate',
        'unit': 'mL/sec/cm3',
    },
    'khe_i': {
        'init': 0.003,
        'default_free': True,
        'bounds': [0.0, 0.1],
        'name': 'Initial hepatocellular uptake rate',
        'unit': 'mL/sec/cm3',
    },
    'khe_f': {
        'init': 0.003,
        'default_free': True,
        'bounds': [0.0, 0.1],
        'name': 'Final hepatocellular uptake rate',
        'unit': 'mL/sec/cm3',
    },
    'Th': {
        'init': 30*60,
        'default_free': True,
        'bounds': [10*60, 10*60*60],
        'name': 'Hepatocellular mean transit time',
        'unit': 'sec',
    },
    'Th_i': {
        'init': 30*60,
        'default_free': True,
        'bounds': [10*60, 10*60*60],
        'name': 'Initial hepatocellular mean transit time',
        'unit': 'sec',
    },
    'Th_f': {
        'init': 30*60,
        'default_free': True,
        'bounds': [10*60, 10*60*60],
        'name': 'Final hepatocellular mean transit time',
        'unit': 'sec',
    },
    'vol': {
        'init': None,
        'default_free': False,
        'bounds': [0, 10000],
        'name': 'Liver volume',
        'unit': 'cm3',
    },
}

PARAMS_DERIVED = {
    'kbh': {
        'name': 'Biliary excretion rate',
        'unit': 'mL/sec/cm3',
    },
    'kbh_i': {
        'name': 'Initial biliary excretion rate',
        'unit': 'mL/sec/cm3',
    },
    'kbh_f': {
        'name': 'Final biliary excretion rate',
        'unit': 'mL/sec/cm3',
    },
    'Kbh': {
        'name': 'Biliary tissue excretion rate',
        'unit': 'mL/sec/cm3',
    },
    'Khe': {
        'name': 'Hepatocellular tissue uptake rate',
        'unit': 'mL/sec/cm3',
    },
    'E': {
        'name': 'Liver extraction fraction',
        'unit': '',
    },
    'Ktrans': {
        'name': 'Hepatic plasma clearance',
        'unit': 'mL/sec/cm3',
    },
    'CL': {
        'name': 'Liver blood clearance',
        'unit': 'mL/sec',
    }, 
}


def _div(a, b):
    with np.errstate(divide='ignore', invalid='ignore'):
        return np.divide(a, b)




import numpy as np

from dcmri import pk, rel, sig



def params_tissue(kinetics='2CX', water_exchange='RR') -> list:
    """Parameters characterizing a 2-site exchange tissue. 
    For more detail see :ref:`two-site-exchange`.

    Args:
        kinetics (str, optional): Tracer-kinetic regime. Possible values are
         '2CX', '2CU', 'HF', 'HFU', 'NX', 'FX', 'WV', 'U'. Defaults to '2CX'.
        water_exchange (str, optional): Water exchange regime. Any combination
          of two of the letters 'F', 'N', 'R' is allowed. Defaults to 'RR'.

    Returns: 
        list: tissue parameters

    Raises:
        ValueError: if the configuration is not recognized.

    Example:

        Print the parameters of a HFU tissue with restricted water 
        exchange:

        >>> import dcmri as dc

        >>> dc.params_tissue('HFU', 'RR')
        ['PSe', 'PSc', 'H', 'vb', 'vi', 'PS']

    Notes:

        Table :ref:`tissue-kinetic-regimes` list the water compartments and free 
        parameters for all configurations. Regimes without water exchange 
        across one or both of the barriers are not listed 
        explicitly (FN, NF, FR, NR and NN). They differ from restricted water 
        exchange only in the sense that the respective water permeabilities 
        *PSe* and/or *PSc* are zero. 

        .. _tissue-kinetic-regimes:
        .. list-table:: **Parameters by configuration** 
            :widths: 15 15 30 40
            :header-rows: 1 

            * - Water exchange
              - Indicator exchange
              - Water compartments
              - Free parameters
            * - **FF**
              - 
              - 
              - 
            * - FF
              - 2CX
              - vb + vi + vc
              - H, vb, vi, Fb, PS
            * - FF
              - 2CU
              - vb + vi + vc
              - H, vb, Fb, PS
            * - FF
              - HF
              - vb + vi + vc
              - H, vb, vi, PS
            * - FF
              - HFU
              - vb + vi + vc
              - H, vb, PS
            * - FF
              - FX
              - vb + vi + vc 
              - H, ve, Fb  
            * - FF
              - NX
              - vb + vi + vc
              - vb, Fb  
            * - FF
              - NXP
              - vb + vi + vc
              - vb, Fb  
            * - FF
              - U
              - vb + vi + vc
              - Fb
            * - FF
              - WV
              - vb + vi + vc
              - H, vi, Ktrans
            * - **RR**
              - 
              - 
              - 
            * - RR
              - 2CX
              - vb, vi, vc
              - PSe, PSc, H, vb, vi, Fb, PS
            * - RR
              - 2CU
              - vb, vi, vc
              - PSe, PSc, H, vb, vi, Fb, PS
            * - RR
              - HF
              - vb, vi, vc
              - PSe, PSc, H, vb, vi, PS
            * - RR
              - HFU
              - vb, vi, vc
              - PSe, PSc, H, vb, vi, PS
            * - RR
              - FX
              - vb, vi, vc 
              - PSe, PSc, H, vb, vi, Fb 
            * - RR
              - NX
              - vb, vi, vc 
              - PSe, vb, vi, Fb  
            * - RR
              - NXP
              - vb, vi, vc 
              - PSe, vb, vi, Fb 
            * - RR
              - U
              - vb, vi, vc 
              - PSe, vb, vi, Fb 
            * - RR
              - WV
              - vi, vi+vc
              - PSc, H, vi, Ktrans
            * - **RF**
              - 
              - 
              - 
            * - RF
              - 2CX
              - vb, vi+vc
              - PSe, H, vb, vi, Fb, PS
            * - RF
              - 2CU
              - vb, vi+vc
              - PSe, H, vb, Fb, PS
            * - RF
              - HF
              - vb, vi+vc
              - PSe, H, vb, vi, PS
            * - RF
              - HFU
              - vb, vi+vc
              - PSe, H, vb, PS
            * - RF
              - FX
              - vb, vi+vc
              - PSe, H, vb, vi, Fb
            * - RF
              - NX
              - vb, vi+vc
              - PSe, vb, Fb
            * - RF
              - NXP
              - vb, vi+vc
              - PSe, vb, Fb
            * - RF
              - U
              - vb, vi+vc
              - PSe, vb, Fb 
            * - RF
              - WV
              - vi+vc
              - H, vi, Ktrans
            * - **FR**
              - 
              - 
              -  
            * - FR
              - 2CX
              - vb+vi, vc
              - PSc, H, vb, vi, Fb, PS
            * - FR
              - 2CU
              - vb+vi, vc
              - PSc, H, vb, vi, Fb, PS
            * - FR
              - HF
              - vb+vi, vc
              - PSc, H, vb, vi, PS
            * - FR
              - HFU
              - vb+vi, vc
              - PSc, H, vb, vi, PS
            * - FR
              - FX
              - vb+vi, vc 
              - PSc, H, vb, vi, Fb
            * - FR
              - NX
              - vb+vi, vc 
              - PSc, vb, vi, Fb
            * - FR
              - NXP
              - vb+vi, vc 
              - PSc, vb, vi, Fb
            * - FR
              - U
              - vb+vi, vc 
              - PSc, vc, Fb
            * - FR
              - WV
              - vi, vc
              - PSc, H, vi, Ktrans
    """

    if kinetics not in ['2CX', '2CU', 'HF', 'HFU', 'WV', 'FX', 'NX', 'NXP', 'U']:
        raise ValueError(
            'The value ' + str(kinetics) + ' for the kinetics argument \
            is not recognised. \n Possible values are 2CX, 2CU, HF, HFU, WV, \
            FX, NX, NXP, U.'
        )

    if water_exchange not in ['FF', 'NF','RF', 'FN', 'NN', 
                              'RN', 'FR', 'NR', 'RR']:
        raise ValueError(
            'The value ' + str(water_exchange) + ' of the \
            water_exchange argument is not recognised.\n It must be a \
            2-element string composed of characters N, F, R.'
        )
    
    pars = _relax_pars(kinetics, water_exchange)

    if water_exchange in ['FF', 'NN', 'NF', 'FN']:
        return pars
    if water_exchange in ['NR', 'FR']:
        return ['PSc'] + pars
    if water_exchange in ['RN', 'RF']:
        if kinetics == 'WV':
            return pars
        else:
            return ['PSe'] + pars
    if water_exchange == 'RR':
        if kinetics == 'WV':
            return ['PSc'] + pars
        else:
            return ['PSe', 'PSc'] + pars


def _relax_pars(kin, wex) -> list:

    if kin == '2CX':
        return ['H', 'vb', 'vi', 'Fb', 'PS']
    if kin == 'HF':
        return ['H', 'vb', 'vi', 'PS']
    if kin == 'WV':
        return ['H', 'vi', 'Ktrans']

    if wex == 'FF':

        if kin == '2CU':
            return ['H', 'vb', 'Fb', 'PS']
        if kin == 'HFU':
            return ['H', 'vb', 'PS']
        if kin == 'FX':
            return ['H', 've', 'Fb']
        if kin == 'NX':
            return ['vb', 'Fb']
        if kin == 'NXP':
            return ['vb', 'Fb']
        if kin == 'U':
            return ['Fb']

    if wex in ['RR', 'NN', 'NR', 'RN']:

        if kin == '2CU':
            return ['H', 'vb', 'vi', 'Fb', 'PS']
        if kin == 'HFU':
            return ['H', 'vb', 'vi', 'PS']
        if kin == 'FX':
            return ['H', 'vb', 'vi', 'Fb']
        if kin == 'NX':
            return ['vb', 'vi', 'Fb']
        if kin == 'NXP':
            return ['vb', 'vi', 'Fb']
        if kin == 'U':
            return ['vb', 'vi', 'Fb']

    if wex in ['RF', 'NF']:

        if kin == '2CU':
            return ['H', 'vb', 'Fb', 'PS']
        if kin == 'HFU':
            return ['H', 'vb', 'PS']
        if kin == 'FX':
            return ['H', 'vb', 'vi', 'Fb']
        if kin == 'NX':
            return ['vb', 'Fb']
        if kin == 'NXP':
            return ['vb', 'Fb']
        if kin == 'U':
            return ['vb', 'Fb']

    if wex in ['FR', 'FN']:

        if kin == '2CU':
            return ['H', 'vb', 'vi', 'Fb', 'PS']
        if kin == 'HFU':
            return ['H', 'vb', 'vi', 'PS']
        if kin == 'FX':
            return ['H', 'vb', 'vi', 'Fb']
        if kin == 'NX':
            return ['vb', 'vi', 'Fb']
        if kin == 'NXP':
            return ['vb', 'vi', 'Fb']
        if kin == 'U':
            return ['vc', 'Fb']


def signal_tissue(
        ca: np.ndarray, R10: float, r1: float, t=None, dt=1.0, 
        kinetics='2CX', water_exchange='FF', sequence=None, inflow=None,
        **params) -> np.ndarray:
    
    """Signal for a 2-site exchange tissue. For more detail see
    :ref:`two-site-exchange`.

    Args:
        ca (array-like): concentration in the blood of the arterial input.
        R10 (float): precontrast relaxation rate. The tissue is assumed to be 
          in fast exchange before injection of contrast agent.
        r1 (float): contrast agent relaxivity. 
        t (array_like, optional): the time points in sec of the input function 
          *ca*. If *t* is not provided, the time points are assumed to be 
          uniformly spaced with spacing *dt*. Defaults to None.
        dt (float, optional): spacing in seconds between time points for 
          uniformly spaced time points. This parameter is ignored if *t* is 
          explicity provided. Defaults to 1.0.
        kinetics (str, optional): Tracer-kinetic model. Possible values are
         '2CX', '2CU', 'HF', 'HFU', 'NX', 'FX', 'WV', 'U'. Defaults to '2CX'.
        water_exchange (str, optional): Water exchange regime, Any combination
          of two of the letters 'F', 'N', 'R' is allowed. Defaults to 'FF'.
        sequence (dict): the sequence model and its parameters. The 
          dictionary has one required key 'model' which specifies the signal 
          model. Currently either 'SS', 'SR' or 'lin'. The other keys are the values 
          of the signal parameter, which depend on the model. See table 
          :ref:`Tissue-signal-parameters` for detail. 
        inflow (dict, optional): inflow model. If not provided, the in- and 
          outflow of magnetization is ignored. To include 
          inflow effects, **inflow** must be dictionary with the signal model 
          parameters for the arterial input. For the 'SS' signal model, 
          required parameters are 'R10a' and 'B1corr_a'. Defaults to None.
        params (dict): model parameters. See :ref:`Tissue-signal-parameters` 
          for more detail. Note: the tissue parameters are keyword 
          arguments for convenience, but a value is required.

    Raises:
        ValueError: if a required parameter has no value assigned.
        NotImplementedError: if a combination of regimes is not yet 
          implemented. Currently the sequence type 'SR' only accepts fast 
          water exchange 'FF'.

    Returns:
        ndarray: tissue signal as a 1D array.  

    Example:

        We verify that the effect of inflow is negligible in a steady state 
        sequence:

    .. plot::
        :include-source:

        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> import dcmri as dc

        Define constants and model parameters: 

        >>> R10, r1 = 1, 5000
        >>> seq = {'model': 'SS', 'S0':1, 'FA':15, 'TR': 0.001, 'B1corr':1, 'noise_sdev':0}
        >>> pars = {
        >>>     'sequence':seq, 'kinetics':'2CX', 'water_exchange':'NN', 
        >>>     'H':0.045, 'vb':0.05, 'vi':0.3, 'Fb':0.01, 'PS':0.005} 
        >>> inflow = {'R10a': 0.7, 'B1corr_a':1}

        Generate arterial blood concentrations:

        >>> t = np.arange(0, 300, 1.5)
        >>> ca = dc.aif_parker(t, BAT=20)/(1-0.45) 

        Calculate the signal with and without inflow:

        >>> Sf = dc.signal_tissue(ca, R10, r1, t=t, inflow=inflow, **pars)
        >>> Sn = dc.signal_tissue(ca, R10, r1, t=t, **pars)

        Compare them in a plot:

        >>> plt.figure()
        >>> plt.plot(t/60, Sn, label='Without inflow correction', linewidth=3)
        >>> plt.plot(t/60, Sf, label='With inflow correction')
        >>> plt.xlabel('Time (min)')
        >>> plt.ylabel('Concentration (mM)')
        >>> plt.legend()
        >>> plt.show()

    Notes:

        .. _Tissue-signal-parameters:
        .. list-table:: **Tissue signal parameters**
            :widths: 20 30 30
            :header-rows: 1

            * - Parameters
              - When to use
              - Further detail
            * - Fb, PS, Ktrans, vb, H, vi,
                ve, vc, PSe, PSc.
              - Depends on **kinetics** and **water_exchange**
              - :ref:`tissue-kinetic-regimes`
            * - S0, noise_sdev
              - Always
              - :ref:`params-per-sequence`
            * - FA, TR, B1corr
              - If **sequence** in ['SR', 'SS']
              - :ref:`params-per-sequence`
            * - TP, TC
              - If **sequence** is 'SR'
              - :ref:`params-per-sequence`
            * - R10a, B1corr_a
              - If **inflow** is not None
              - :ref:`relaxation-params`, :ref:`params-per-sequence`

    """
    if sequence is None:
        raise ValueError(
            'sequence is required. Please specify a model and appropriate '
            'sequence parameters.')
    
    R1, v, Fw = relax_tissue(
        ca, R10, r1, t=t, dt=dt, 
        kinetics=kinetics, water_exchange=water_exchange, **params)
    
    if sequence['model'] == 'SS':
        if inflow is None:
            j = None
        else:
            if kinetics != '2CX':
                raise ValueError('Inflow correction is currently only \
                                 available for 2CX tissues')
            FAa = inflow['B1corr_a'] * sequence['FA']
            R1a = rel.relax(ca, inflow['R10a'], r1)
            na = sig.Mz_ss(R1a, sequence['TR'], FAa)
            if np.isscalar(v):
                j = Fw*na
            else:
                j = np.zeros((len(v), len(na)))
                j[0, :] = Fw[0,0]*na
        FA = sequence['B1corr'] * sequence['FA']
        return sig.signal_ss(
            sequence['S0'], R1, sequence['TR'], FA, v, Fw, j, noise_sdev=sequence['noise_sdev'])
    
    elif sequence['model'] == 'SR':
        if inflow is None:
            j = None
        else:
            if kinetics != '2CX':
                raise ValueError('Inflow correction is currently only \
                                 available for 2CX tissues')
            FAa = inflow['B1corr_a'] * sequence['FA']
            R1a = rel.relax(ca, inflow['R10a'], r1)
            na = sig.Mz_spgr(
                R1a, sequence['TC'], sequence['TR'], FAa, sequence['TP'])
            if np.isscalar(v):
                j = Fw*na
            else:
                j = np.zeros((len(v), len(na)))
                j[0, :] = Fw[0,0]*na
        FA = sequence['B1corr'] * sequence['FA']
        return sig.signal_spgr(
            sequence['S0'], R1, sequence['TC'], sequence['TR'], FA, 
            sequence['TP'], v=v, Fw=Fw, noise_sdev=sequence['noise_sdev'])
    

def Mz_tissue(
        ca: np.ndarray, R10: float, r1: float, t=None, dt=1.0, 
        kinetics='2CX', water_exchange='FF', sequence=None, inflow=None,
        **params) -> np.ndarray:
    
    """Longitudinal magnetization for a 2-site exchange tissue. For more 
    detail see :ref:`two-site-exchange`.

    Args:
        ca (array-like): concentration in the blood of the arterial input.
        R10 (float): precontrast relaxation rate. The tissue is assumed to be 
          in fast exchange before injection of contrast agent.
        r1 (float): contrast agent relaxivity. 
        t (array_like, optional): the time points in sec of the input function 
          *ca*. If *t* is not provided, the time points are assumed to be 
          uniformly spaced with spacing *dt*. Defaults to None.
        dt (float, optional): spacing in seconds between time points for 
          uniformly spaced time points. This parameter is ignored if *t* is 
          explicity provided. Defaults to 1.0.
        kinetics (str, optional): Tracer-kinetic model. Possible values are
         '2CX', '2CU', 'HF', 'HFU', 'NX', 'FX', 'WV', 'U'. Defaults to '2CX'.
        water_exchange (str, optional): Water exchange regime, Any combination
          of two of the letters 'F', 'N', 'R' is allowed. Defaults to 'FF'.
        sequence (dict): the sequence model and its parameters. The 
          dictionary has one required key 'model' which specifies the signal 
          model. Currently either 'SS' or 'SR'. The other keys are the values 
          of the signal parameter, which depend on the model. See table 
          :ref:`Tissue-signal-parameters` for detail. 
        inflow (dict, optional): inflow model. If not provided, the in- and 
          outflow of magnetization is ignored. To include 
          inflow effects, **inflow** must be dictionary with the signal model 
          parameters for the arterial input. For the 'SS' signal model, 
          required parameters are 'R10a' and 'B1corr_a'. Defaults to None.
        params (dict): model parameters. See :ref:`Tissue-signal-parameters` 
          for more detail. Note: the tissue parameters are keyword 
          arguments for convenience, but a value is required.

    Raises:
        ValueError: if a required parameter has no value assigned.
        NotImplementedError: if a combination of regimes is not yet 
          implemented. Currently the sequence type 'SR' only accepts fast 
          water exchange 'FF'.

    Returns:
        ndarray: tissue signal as a 1D array.  

    Example:

        We verify that the effect of inflow is negligible in a steady state 
        sequence:

    .. plot::
        :include-source:

        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> import dcmri as dc

        Define constants and model parameters: 

        >>> R10, r1 = 1, 5000
        >>> seq = {'model': 'SS', 'FA':15, 'TR': 0.001, 'B1corr':1}
        >>> pars = {
        >>>     'sequence':seq, 'kinetics':'2CX', 'water_exchange':'NN', 
        >>>     'H':0.045, 'vb':0.05, 'vi':0.3, 'Fb':0.01, 'PS':0.005} 
        >>> inflow = {'R10a': 0.7, 'B1corr_a':1}

        Generate arterial blood concentrations:

        >>> t = np.arange(0, 300, 1.5)
        >>> ca = dc.aif_parker(t, BAT=20)/(1-0.45) 

        Calculate the signal with and without inflow:

        >>> Mf = dc.Mz_tissue(ca, R10, r1, t=t, inflow=inflow, **pars)
        >>> Mn = dc.Mz_tissue(ca, R10, r1, t=t, **pars)

        Compare them in a plot:

        >>> plt.figure()
        >>> plt.plot(t/60, np.sum(Mn, axis=0), label='No inflow', linewidth=3)
        >>> plt.plot(t/60, np.sum(Mf, axis=0), label='Inflow')
        >>> plt.xlabel('Time (min)')
        >>> plt.ylabel('Magnetization (A/cm)')
        >>> plt.legend()
        >>> plt.show()

    Notes:

        .. _Mz-signal-parameters:
        .. list-table:: **Tissue Mz parameters**
            :widths: 20 30 30
            :header-rows: 1

            * - Parameters
              - When to use
              - Further detail
            * - Fb, PS, Ktrans, vb, H, vi,
                ve, vc, PSe, PSc.
              - Depends on **kinetics** and **water_exchange**
              - :ref:`tissue-kinetic-regimes`
            * - FA, TR, B1corr
              - Always
              - :ref:`params-per-sequence`
            * - TP, TC
              - If **sequence** is 'SR'
              - :ref:`params-per-sequence`
            * - R10a, B1corr_a
              - If **inflow** is not None
              - :ref:`relaxation-params`, :ref:`params-per-sequence`

    """
    if sequence is None:
        raise ValueError(
            'sequence is required. Please specify a model \
             and appropriate sequence parameters.')
    
    R1, v, Fw = relax_tissue(
        ca, R10, r1, t=t, dt=dt, 
        kinetics=kinetics, water_exchange=water_exchange, **params)
    
    if sequence['model'] == 'SS':
        if inflow is None:
            j = None
        else:
            if kinetics != '2CX':
                raise ValueError('Inflow correction is currently only '
                                 'available for 2CX tissues')
            FAa = inflow['B1corr_a'] * sequence['FA']
            R1a = rel.relax(ca, inflow['R10a'], r1)
            na = sig.Mz_ss(R1a, sequence['TR'], FAa)
            if np.isscalar(v):
                j = Fw*na
            else:
                j = np.zeros((len(v), len(na)))
                j[0, :] = Fw[0,0]*na
        FA = sequence['B1corr'] * sequence['FA']
        return sig.Mz_ss(R1, sequence['TR'], FA, v, Fw, j)
    
    elif sequence['model'] == 'SR':
        if inflow is None:
            j = None
        else:
            if kinetics != '2CX':
                raise ValueError('Inflow correction is currently only '
                                 'available for 2CX tissues')
            FAa = inflow['B1corr_a'] * sequence['FA']
            R1a = rel.relax(ca, inflow['R10a'], r1)
            na = sig.Mz_spgr(
                R1a, sequence['TC'], sequence['TR'], FAa, sequence['TP'])
            if np.isscalar(v):
                j = Fw*na
            else:
                j = np.zeros((len(v), len(na)))
                j[0, :] = Fw[0,0]*na
        FA = sequence['B1corr'] * sequence['FA']
        return sig.Mz_spgr(
            R1, sequence['TC'], sequence['TR'], FA,  sequence['TP'], v, Fw)
    

def relax_tissue(ca: np.ndarray, R10: float, r1: float, t=None, dt=1.0, 
                 kinetics='2CX', water_exchange='FF', **params):
    """Free relaxation rates for a 2-site exchange tissue. For more detail see
    :ref:`two-site-exchange`.

    Note: the free relaxation rates are the relaxation rates of the tissue 
    compartments in the absence of water exchange between them.

    Args:
        ca (array-like): concentration in the blood of the arterial input.
        R10 (float): precontrast relaxation rate. The tissue is assumed to be 
          in fast exchange before injection of contrast agent.
        r1 (float): contrast agent relaxivity. 
        t (array_like, optional): the time points in sec of the input function 
          *ca*. If *t* is not provided, the time points are assumed to be 
          uniformly spaced with spacing *dt*. Defaults to None.
        dt (float, optional): spacing in seconds between time points for 
          uniformly spaced time points. This parameter is ignored if *t* is 
          explicity provided. Defaults to 1.0.
        kinetics (str, optional): Tracer-kinetic model. Possible values are
         '2CX', '2CU', 'HF', 'HFU', 'NX', 'FX', 'WV', 'U'. Defaults to '2CX'.
        water_exchange (str, optional): Water exchange regime, Any combination
          of two of the letters 'F', 'N', 'R' is allowed. Defaults to 'FF'.
        params (dict): values for the parameters of the tissue,
          specified as keyword parameters. See table :ref:`tissue-kinetic-regimes` 
          for more detail on the parameters that are relevant in each regime. 

    Returns:
        numpy.ndarray: relaxation rates
          In the fast water exchange limit, the 
          relaxation rates are a 1D array. In all other situations, 
          relaxation rates are a 2D-array with dimensions (k,n), where k is 
          the number of compartments and n is the number of time points 
          in ca.
        volume fractions
          the volume fractions of the tissue compartments. 
          Returns None in 'FF' regime. 
        water flows
          2D array with water exchange 
          rates between tissue compartments. Returns None in 'FF' regime.

    Example:

        Compare the free relaxation rates without water exchange against 
        relaxation rates in fast exchange:

    .. plot::
        :include-source:

        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> import dcmri as dc

        Generate a population-average input function:

        >>> t = np.arange(0, 300, 1.5)
        >>> ca = dc.aif_parker(t, BAT=20)

        Define constants and model parameters: 

        >>> R10, r1 = 1/dc.T1(), dc.relaxivity()     
        >>> pf = {'H':0.5, 'vb':0.05, 'vi':0.3, 'Fb':0.01, 'PS':0.005}   
        >>> pn = {'H':0.5, 'vb':0.1, 'vi':0.3, 'Fb':0.01, 'PS':0.005}

        Calculate tissue relaxation rates without water exchange, 
        and also in the fast exchange limit for comparison:

        >>> R1f, _, _ = dc.relax_tissue(ca, R10, r1, t=t, water_exchange='FF', **pf)
        >>> R1n, _, _ = dc.relax_tissue(ca, R10, r1, t=t, water_exchange='NN', **pn)

        Plot the relaxation rates in the three compartments, and compare 
        against the fast exchange result:

        >>> fig, (ax0, ax1) = plt.subplots(1,2,figsize=(12,5))

        Plot restricted water exchange in the left panel:

        >>> ax0.set_title('Restricted water exchange')
        >>> ax0.plot(t/60, R1n[0,:], linestyle='-', 
        >>>          linewidth=2.0, color='darkred', label='Blood')
        >>> ax0.plot(t/60, R1n[1,:], linestyle='-', 
        >>>          linewidth=2.0, color='darkblue', label='Interstitium')
        >>> ax0.plot(t/60, R1n[2,:], linestyle='-', 
        >>>          linewidth=2.0, color='grey', label='Cells')
        >>> ax0.set_xlabel('Time (min)')
        >>> ax0.set_ylabel('Compartment relaxation rate (1/sec)')
        >>> ax0.legend()

        Plot fast water exchange in the right panel:

        >>> ax1.set_title('Fast water exchange')
        >>> ax1.plot(t/60, R1f, linestyle='-', 
        >>>          linewidth=2.0, color='black', label='Tissue')
        >>> ax1.set_xlabel('Time (min)')
        >>> ax1.set_ylabel('Tissue relaxation rate (1/sec)')
        >>> ax1.legend()
        >>> plt.show()

    """

    # Check configuration
    if kinetics not in ['U', 'FX', 'NX', 'NXP', 'WV', 'HFU', 'HF', '2CU', '2CX']:
        raise ValueError(
            "Kinetic model '" + str(kinetics) + 
            "' is not recognised.\n Possible values are: " +
            "'U', 'FX', 'NX', 'NXP', 'WV', 'HFU', 'HF', '2CU' and '2CX'."
        )
    
    if (water_exchange[0] not in ['F', 'R', 'N']) or (
        water_exchange[1] not in ['F', 'R', 'N']):
        raise ValueError(
            "Water exchange regime '" +
            str(water_exchange) + "' is not recognised.\n" + 
            "Possible values are: 'FF','RF','NF','FR','RR','NR','FN','RN','NN'"
        )

    req = params_tissue(kinetics, water_exchange)
    if set(req) != set(params.keys()):
        raise ValueError(
            "Model parameters are incorrect or incomplete. A model with "
            "kinetics " + kinetics + " and water_exchange "+water_exchange+" requires "
            "parameters " + str(req) + " but you have provided "+str(list(params.keys())))

    if water_exchange[0] == 'N':
        if kinetics != 'WV':
            params['PSe'] = 0

    if water_exchange[1] == 'N':
        params['PSc'] = 0    

    wex = water_exchange.replace('N','R')

    # Distribute cases
    if wex == 'FF':

        if kinetics == 'U':
            return _relax_u_ff(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'FX':
            return _relax_fx_ff(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'NX':
            return _relax_nx_ff(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'NXP':
            return _relax_nxp_ff(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'WV':
            return _relax_wv_ff(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'HFU':
            return _relax_hfu_ff(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'HF':
            return _relax_hf_ff(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == '2CU':
            return _relax_2cu_ff(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == '2CX':
            return _relax_2cx_ff(ca, R10, r1, t=t, dt=dt, **params)

    elif wex == 'RF':

        if kinetics == 'U':
            return _relax_u_rf(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'FX':
            return _relax_fx_rf(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'NX':
            return _relax_nx_rf(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'NXP':
            return _relax_nxp_rf(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'WV':
            return _relax_wv_rf(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'HFU':
            return _relax_hfu_rf(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'HF':
            return _relax_hf_rf(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == '2CU':
            return _relax_2cu_rf(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == '2CX':
            return _relax_2cx_rf(ca, R10, r1, t=t, dt=dt, **params)

    elif wex == 'FR':

        if kinetics == 'U':
            return _relax_u_fr(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'FX':
            return _relax_fx_fr(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'NX':
            return _relax_nx_fr(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'NXP':
            return _relax_nxp_fr(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'WV':
            return _relax_wv_fr(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'HFU':
            return _relax_hfu_fr(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'HF':
            return _relax_hf_fr(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == '2CU':
            return _relax_2cu_fr(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == '2CX':
            return _relax_2cx_fr(ca, R10, r1, t=t, dt=dt, **params)

    elif wex == 'RR':

        if kinetics == 'U':
            return _relax_u_rr(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'FX':
            return _relax_fx_rr(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'NX':
            return _relax_nx_rr(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'NXP':
            return _relax_nxp_rr(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'WV':
            return _relax_wv_rr(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'HFU':
            return _relax_hfu_rr(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == 'HF':
            return _relax_hf_rr(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == '2CU':
            return _relax_2cu_rr(ca, R10, r1, t=t, dt=dt, **params)
        elif kinetics == '2CX':
            return _relax_2cx_rr(ca, R10, r1, t=t, dt=dt, **params)


def _c(C,v):
    if v==0:
        # In this case the result does not matter
        return C*0
    else:
        return C/v
    

# FF

# For water flow modelling
def _relax_2cx_ff(ca, R10, r1, t=None, dt=1.0, 
                  H=None, vi=None, vb=None, Fb=None, PS=None):
    C = _conc_2cx(ca, t=t, dt=dt, vi=vi, H=H, vb=vb, Fb=Fb, PS=PS)
    R1 = rel.relax(C, R10, r1)
    return R1, 1, Fb

def _relax_2cu_ff(ca, R10, r1, t=None, dt=1.0, 
                  H=None, vb=None, Fb=None, PS=None):
    C = _conc_2cu(ca, t=t, dt=dt, H=H, vb=vb, Fb=Fb, PS=PS)
    R1 = rel.relax(C, R10, r1)
    return R1, 1, 0

def _relax_hf_ff(ca, R10, r1, t=None, dt=1.0, 
                 H=None, vi=None, vb=None, PS=None):
    C = _conc_hf(ca, t=t, dt=dt, H=H, vb=vb, vi=vi, PS=PS)
    R1 = rel.relax(C, R10, r1)
    return R1, 1, 0

def _relax_hfu_ff(ca, R10, r1, t=None, dt=1.0, 
                 H=None, vb=None, PS=None):
    C = _conc_hfu(ca, t=t, dt=dt, H=H, vb=vb, PS=PS)
    R1 = rel.relax(C, R10, r1)
    return R1, 1, 0

def _relax_nx_ff(ca, R10, r1, t=None, dt=1.0, 
                 vb=None, Fb=None):
    C = _conc_nx(ca, t=t, dt=dt, vb=vb, Fb=Fb)
    R1 = rel.relax(C, R10, r1)
    return R1, 1, 0

def _relax_nxp_ff(ca, R10, r1, t=None, dt=1.0, 
                 vb=None, Fb=None):
    C = _conc_nxp(ca, t=t, dt=dt, vb=vb, Fb=Fb)
    R1 = rel.relax(C, R10, r1)
    return R1, 1, 0

def _relax_nxp_ff(ca, R10, r1, t=None, dt=1.0, 
                 vb=None, Fb=None):
    C = _conc_nxp(ca, t=t, dt=dt, vb=vb, Fb=Fb)
    R1 = rel.relax(C, R10, r1)
    return R1, 1, 0

def _relax_wv_ff(ca, R10, r1, t=None, dt=1.0,
                 H=None, vi=None, Ktrans=None):
    C = _conc_wv(ca, t=t, dt=dt, H=H, vi=vi, Ktrans=Ktrans)
    R1 = rel.relax(C, R10, r1)
    return R1, 1, 0

def _relax_u_ff(ca, R10, r1, t=None, dt=1.0, 
                Fb=None):
    C = _conc_u(ca, t=t, dt=dt, Fb=Fb)
    R1 = rel.relax(C, R10, r1)
    return R1, 1, 0

def _relax_fx_ff(ca, R10, r1, t=None, dt=1.0,
                 H=None, ve=None, Fb=None):
    C = _conc_fx(ca, t=t, dt=dt, H=H, ve=ve, Fb=Fb)
    R1 = rel.relax(C, R10, r1)
    return R1, 1, 0

# FR

# For water flow modelling
def _relax_2cx_fr(ca, R10, r1, t=None, dt=1.0, 
                  H=None, vb=None, vi=None, Fb=None, PS=None, PSc=None):
    C = _conc_2cx(ca, t=t, dt=dt, H=H, vb=vb, vi=vi, Fb=Fb, PS=PS)
    v = [vb+vi, 1-vb-vi]
    R1 = (
        rel.relax(_c(C, v[0]), R10, r1),
        rel.relax(ca*0, R10, r1), 
    )
    Fw = [[Fb, PSc], [PSc, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

def _relax_2cu_fr(ca, R10, r1, t=None, dt=1.0, 
                  H=None, vi=None, vb=None, Fb=None, PS=None, PSc=None):
    C = _conc_2cu(ca, t=t, dt=dt, H=H, vb=vb, Fb=Fb, PS=PS)
    vc = 1-vb-vi
    v = [1-vc, vc]
    R1 = (
        rel.relax(_c(C, v[0]), R10, r1),
        rel.relax(ca*0, R10, r1), 
    )
    Fw = [[0, PSc], [PSc, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

def _relax_hf_fr(ca, R10, r1, t=None, dt=1.0, 
                 H=None, vb=None, vi=None, PS=None, PSc=None):
    C = _conc_hf(ca, t=t, dt=dt, H=H, vb=vb, vi=vi, PS=PS)
    v = [vb+vi, 1-vb-vi]
    R1 = (
        rel.relax(_c(C, v[0]), R10, r1),
        rel.relax(ca*0, R10, r1), 
    )
    Fw = [[0, PSc], [PSc, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

def _relax_hfu_fr(ca, R10, r1, t=None, dt=1.0, 
                 H=None, vi=None, vb=None, PS=None, PSc=None):
    C = _conc_hfu(ca, t=t, dt=dt, H=H, vb=vb, PS=PS)
    vc = 1-vb-vi
    v = [1-vc, vc]
    R1 = (
        rel.relax(_c(C, v[0]), R10, r1),
        rel.relax(ca*0, R10, r1), 
    )
    Fw = [[0, PSc], [PSc, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

def _relax_wv_fr(ca, R10, r1, t=None, dt=1.0, 
                 H=None, vi=None, Ktrans=None, PSc=None):
    C = _conc_wv(ca, t=t, dt=dt, H=H, vi=vi, Ktrans=Ktrans)
    v = [vi, 1-vi]
    R1 = (
        rel.relax(_c(C, v[0]), R10, r1),
        rel.relax(ca*0, R10, r1), 
    )
    Fw = [[0, PSc], [PSc, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

def _relax_fx_fr(ca, R10, r1, t=None, dt=1.0, 
                 H=None, vb=None, vi=None, Fb=None, PSc=None):
    vp = vb*(1-H)
    ve = vi+vp
    vc = 1-vb-vi
    C = _conc_fx(ca, t=t, dt=dt, H=H, ve=ve, Fb=Fb)
    v = [1-vc, vc]
    R1 = (
        rel.relax(_c(C, v[0]), R10, r1),
        rel.relax(ca*0, R10, r1), 
    )
    Fw = [[0, PSc], [PSc, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

def _relax_nx_fr(ca, R10, r1, t=None, dt=1.0, 
                 vi=None, vb=None, Fb=None, PSc=None):
    vc = 1-vb-vi
    C = _conc_nx(ca, t=t, dt=dt, vb=vb, Fb=Fb)
    v = [1-vc, vc]
    R1 = (
        rel.relax(_c(C, v[0]), R10, r1),
        rel.relax(ca*0, R10, r1), 
    )
    Fw = [[0, PSc], [PSc, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

def _relax_nxp_fr(ca, R10, r1, t=None, dt=1.0, 
                 vi=None, vb=None, Fb=None, PSc=None):
    vc = 1-vb-vi
    C = _conc_nxp(ca, t=t, dt=dt, vb=vb, Fb=Fb)
    v = [1-vc, vc]
    R1 = (
        rel.relax(_c(C, v[0]), R10, r1),
        rel.relax(ca*0, R10, r1), 
    )
    Fw = [[0, PSc], [PSc, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

def _relax_u_fr(ca, R10, r1, t=None, dt=1.0, 
                vc=None, Fb=None, PSc=None):
    C = _conc_u(ca, t=t, dt=dt, Fb=Fb)
    v = [1-vc, vc]
    R1 = (
        rel.relax(_c(C, v[0]), R10, r1),
        rel.relax(ca*0, R10, r1), 
    )
    Fw = [[0, PSc], [PSc, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

# RF

# For water flow modelling
def _relax_2cx_rf(ca, R10, r1, t=None, dt=1.0, 
                  H=None, vb=None, vi=None, Fb=None, PS=None, PSe=None):
    C = _conc_2cx(ca, t=t, dt=dt, sum=False, 
                  H=H, vb=vb, vi=vi, Fb=Fb, PS=PS)
    v = [vb, 1-vb]
    R1 = (
        rel.relax(_c(C[0,:], v[0]), R10, r1),
        rel.relax(_c(C[1,:], v[1]), R10, r1),
    )
    Fw = [[Fb, PSe], [PSe, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)


def _relax_2cu_rf(ca, R10, r1, t=None, dt=1.0, 
                  H=None, vb=None, Fb=None, PS=None, PSe=None):
    C = _conc_2cu(ca, t=t, dt=dt, sum=False, 
                  H=H, vb=vb, Fb=Fb, PS=PS)
    v = [vb, 1-vb]
    R1 = (
        rel.relax(_c(C[0,:], v[0]), R10, r1),
        rel.relax(_c(C[1,:], v[1]), R10, r1),
    )
    Fw = [[0, PSe], [PSe, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

def _relax_hf_rf(ca, R10, r1, t=None, dt=1.0, 
                 H=None, vb=None, vi=None, PS=None, PSe=None):
    C = _conc_hf(ca, t=t, dt=dt, sum=False, H=H, vb=vb, vi=vi, PS=PS)
    v = [vb, 1-vb]
    R1 = (
        rel.relax(_c(C[0,:], v[0]), R10, r1),
        rel.relax(_c(C[1,:], v[1]), R10, r1),
    )
    Fw = [[0, PSe], [PSe, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

def _relax_hfu_rf(ca, R10, r1, t=None, dt=1.0, 
                 H=None, vb=None, PS=None, PSe=None):
    C = _conc_hfu(ca, t=t, dt=dt, sum=False, H=H, vb=vb, PS=PS)
    v = [vb, 1-vb]
    R1 = (
        rel.relax(_c(C[0,:], v[0]), R10, r1),
        rel.relax(_c(C[1,:], v[1]), R10, r1),
    )
    Fw = [[0, PSe], [PSe, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

def _relax_wv_rf(ca, R10, r1, t=None, dt=1.0, 
                 H=None, vi=None, Ktrans=None):
    C = _conc_wv(ca, t=t, dt=dt, H=H, vi=vi, Ktrans=Ktrans)
    R1 = rel.relax(C, R10, r1)
    return R1, 1, 0

def _relax_fx_rf(ca, R10, r1, t=None, dt=1.0, 
                 H=None, vb=None, vi=None, Fb=None, PSe=None):
    vp = vb * (1-H)
    ve = vp + vi
    C = _conc_fx(ca, t=t, dt=dt, H=H, ve=ve, Fb=Fb)
    v = [vb, 1-vb]
    if ve==0:
        Cp = C*0
        Ci = C*0
    else:
        Cp = C*vp/ve
        Ci = C*vi/ve
    R1 = (
        rel.relax(_c(Cp, v[0]), R10, r1),
        rel.relax(_c(Ci, v[1]), R10, r1),
    )
    Fw = [[0, PSe], [PSe, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

def _relax_nx_rf(ca, R10, r1, t=None, dt=1.0, 
                 vb=None, Fb=None, PSe=None):
    C = _conc_nx(ca, t=t, dt=dt, vb=vb, Fb=Fb)
    v = [vb, 1-vb]
    R1 = (
        rel.relax(_c(C, v[0]), R10, r1),
        rel.relax(ca*0, R10, r1),
    )
    Fw = [[0, PSe], [PSe, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

def _relax_nxp_rf(ca, R10, r1, t=None, dt=1.0, 
                 vb=None, Fb=None, PSe=None):
    C = _conc_nxp(ca, t=t, dt=dt, vb=vb, Fb=Fb)
    v = [vb, 1-vb]
    R1 = (
        rel.relax(_c(C, v[0]), R10, r1),
        rel.relax(ca*0, R10, r1),
    )
    Fw = [[0, PSe], [PSe, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

def _relax_u_rf(ca, R10, r1, t=None, dt=1.0, 
                vb=None, Fb=None, PSe=None):
    C = _conc_u(ca, t=t, dt=dt, Fb=Fb)
    v = [vb, 1-vb]
    R1 = (
        rel.relax(_c(C, v[0]), R10, r1),
        rel.relax(ca*0, R10, r1),
    )
    Fw = [[0, PSe], [PSe, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)


# RR

# For water flow modelling
def _relax_2cx_rr(ca, R10, r1, t=None, dt=1.0, 
                  H=None, vb=None, vi=None, 
                  Fb=None, PS=None, PSe=None, PSc=None):
    C = _conc_2cx(ca, t=t, dt=dt, sum=False, 
                  H=H, vb=vb, vi=vi, Fb=Fb, PS=PS)
    v = [vb, vi, 1-vb-vi]
    R1 = (
        rel.relax(_c(C[0,:], v[0]), R10, r1),
        rel.relax(_c(C[1,:], v[1]), R10, r1),
        rel.relax(ca*0, R10, r1),
    )
    Fw = [[Fb, PSe, 0], [PSe, 0, PSc], [0, PSc, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)


def _relax_2cu_rr(ca, R10, r1, t=None, dt=1.0, 
                  H=None, vb=None, vi=None, 
                  Fb=None, PS=None, PSe=None, PSc=None):
    C = _conc_2cu(ca, t=t, dt=dt, sum=False, H=H, vb=vb, Fb=Fb, PS=PS)
    v = [vb, vi, 1-vb-vi]
    R1 = (
        rel.relax(_c(C[0,:], v[0]), R10, r1),
        rel.relax(_c(C[1,:], v[1]), R10, r1), 
        rel.relax(ca*0, R10, r1), 
    )
    Fw = [[0, PSe, 0], [PSe, 0, PSc], [0, PSc, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

def _relax_hf_rr(ca, R10, r1, t=None, dt=1.0, 
                 H=None, vb=None, vi=None, PS=None, PSe=None, PSc=None):
    C = _conc_hf(ca, t=t, dt=dt, sum=False, H=H, vb=vb, vi=vi, PS=PS)
    v = [vb, vi, 1-vb-vi]
    R1 = (
        rel.relax(_c(C[0,:], v[0]), R10, r1),
        rel.relax(_c(C[1,:], v[1]), R10, r1), 
        rel.relax(ca*0, R10, r1), 
    )
    Fw = [[0, PSe, 0], [PSe, 0, PSc], [0, PSc, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

def _relax_hfu_rr(ca, R10, r1, t=None, dt=1.0, 
                  H=None, vb=None, vi=None, PS=None, 
                  PSe=None, PSc=None):
    C = _conc_hfu(ca, t=t, dt=dt, sum=False, H=H, vb=vb, PS=PS)
    v = [vb, vi, 1-vb-vi]
    R1 = (
        rel.relax(_c(C[0,:], v[0]), R10, r1),
        rel.relax(_c(C[1,:], v[1]), R10, r1), 
        rel.relax(ca*0, R10, r1), 
    )
    Fw = [[0, PSe, 0], [PSe, 0, PSc], [0, PSc, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

def _relax_wv_rr(ca, R10, r1, t=None, dt=1.0, 
                 H=None, vi=None, Ktrans=None, PSc=None):
    C = _conc_wv(ca, t=t, dt=dt, H=H, vi=vi, Ktrans=Ktrans)
    v = [vi, 1-vi]
    R1 = (
        rel.relax(_c(C, v[0]), R10, r1),
        rel.relax(ca*0, R10, r1), 
    )
    Fw = [[0, PSc], [PSc, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

def _relax_fx_rr(ca, R10, r1, t=None, dt=1.0, 
                 H=None, vb=None, vi=None, Fb=None, 
                 PSe=None, PSc=None):
    vp = vb * (1-H)
    ve = vp + vi
    C = _conc_fx(ca, t=t, dt=dt, H=H, ve=ve, Fb=Fb)
    v = [vb, vi, 1-vb-vi]
    if ve==0:
        Cp = C*0
        Ci = C*0
    else:
        Cp = C*vp/ve
        Ci = C*vi/ve
    R1 = (
        rel.relax(_c(Cp, v[0]), R10, r1),
        rel.relax(_c(Ci, v[1]), R10, r1), 
        rel.relax(ca*0, R10, r1), 
    )
    Fw = [[0, PSe, 0], [PSe, 0, PSc], [0, PSc, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

def _relax_nx_rr(ca, R10, r1, t=None, dt=1.0, 
                 vb=None, vi=None, Fb=None, 
                 PSe=None, PSc=None):
    C = _conc_nx(ca, t=t, dt=dt, vb=vb, Fb=Fb)
    v = [vb, vi, 1-vb-vi]
    R1 = (
        rel.relax(_c(C, v[0]), R10, r1),
        rel.relax(ca*0, R10, r1),
        rel.relax(ca*0, R10, r1), 
    )
    Fw = [[0, PSe, 0], [PSe, 0, PSc], [0, PSc, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

def _relax_nxp_rr(ca, R10, r1, t=None, dt=1.0, 
                 vb=None, vi=None, Fb=None, 
                 PSe=None, PSc=None):
    C = _conc_nxp(ca, t=t, dt=dt, vb=vb, Fb=Fb)
    v = [vb, vi, 1-vb-vi]
    R1 = (
        rel.relax(_c(C, v[0]), R10, r1),
        rel.relax(ca*0, R10, r1),
        rel.relax(ca*0, R10, r1), 
    )
    Fw = [[0, PSe, 0], [PSe, 0, PSc], [0, PSc, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)

def _relax_u_rr(ca, R10, r1, t=None, dt=1.0, 
                vb=None, vi=None, Fb=None, PSe=None, PSc=None):
    v = [vb, vi, 1-vb-vi]
    if Fb==0:
        R1 = (
            rel.relax(ca*0, R10, r1),
            rel.relax(ca*0, R10, r1), 
            rel.relax(ca*0, R10, r1), 
        )
    else:
        C = _conc_u(ca, t=t, dt=dt, Fb=Fb)
        R1 = (
            rel.relax(_c(C, v[0]), R10, r1),
            rel.relax(ca*0, R10, r1), 
            rel.relax(ca*0, R10, r1), 
        )
    Fw = [[0, PSe, 0], [PSe, 0, PSc], [0, PSc, 0]]
    return np.stack(R1), np.array(v), np.array(Fw)



def conc_tissue(ca: np.ndarray, t=None, dt=1.0, kinetics='2CX', sum=True, 
                **params) -> np.ndarray:
    """Tissue concentration in a 2-site exchange tissue.

    Args:
        ca (array-like): concentration in the arterial input.
        t (array_like, optional): the time points of the input function *ca*. 
          If *t* is not provided, the time points are assumed to be uniformly 
          spaced with spacing *dt*. Defaults to None.
        dt (float, optional): spacing in seconds between time points for 
          uniformly spaced time points. This parameter is ignored if *t* is 
          provided. Defaults to 1.0.
        kinetics (str, optional): Tracer-kinetic model. Possible values are 
          '2CX', '2CU', 'HF', 'HFU', 'NX', 'FX', 'WV', 'U' (see 
          table :ref:`two-site-exchange-kinetics` for detail). Defaults to 
          '2CX'.
        sum (bool, optional): For two-compartment tissues, set to True to 
          return the total tissue concentration, and False to return the 
          concentrations in the compartments separately. In one-compartment 
          tissues this keyword has no effect. Defaults to True.
        params (dict): free model parameters provided as keyword arguments. 
          Possible parameters depend on **kinetics** as detailed in Table 
          :ref:`two-site-exchange-kinetics`. 

    Returns:
        numpy.ndarray: concentration
          If sum=True, or the tissue is one-compartmental, this 
          is a 1D array with the total concentration at each time point. If 
          sum=False this is the concentration in each compartment, and at each 
          time point, as a 2D array with dimensions *(2,k)*, where *k* is the 
          number of time points in *ca*. 

    Raises:
        ValueError: if values are not provided for one or more of the model 
          parameters.

    Example:

        We plot the concentrations of 2CX and WV models with the same values 
        for the shared tissue parameters. 

    .. plot::
        :include-source:

        Start by importing the packages:

        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> import dcmri as dc

        Generate a population-average input function:

        >>> t = np.arange(0, 300, 1.5)
        >>> ca = dc.aif_parker(t, BAT=20)

        Define some tissue parameters: 

        >>> p2x = {'H': 0.5, 'vb':0.1, 'vi':0.4, 'Fb':0.02, 'PS':0.005}
        >>> pwv = {'H': 0.5, 'vi':0.4, 'Ktrans':0.005*0.01/(0.005+0.01)}

        Generate plasma and extravascular tissue concentrations with the 2CX 
        and WV models:

        >>> C2x = dc.conc_tissue(ca, t=t, sum=False, kinetics='2CX', **p2x)
        >>> Cwv = dc.conc_tissue(ca, t=t, kinetics='WV', **pwv)

        Compare them in a plot:

        >>> fig, (ax0, ax1) = plt.subplots(1,2,figsize=(12,5))

        Plot 2CX results in the left panel:

        >>> ax0.set_title('2-compartment exchange model')
        >>> ax0.plot(t/60, 1000*C2x[0,:], linestyle='-', linewidth=3.0, 
        >>>          color='darkred', label='Plasma')
        >>> ax0.plot(t/60, 1000*C2x[1,:], linestyle='-', linewidth=3.0, 
        >>>          color='darkblue', 
        >>>          label='Extravascular, extracellular space')
        >>> ax0.plot(t/60, 1000*(C2x[0,:]+C2x[1,:]), linestyle='-', 
        >>>          linewidth=3.0, color='grey', label='Tissue')
        >>> ax0.set_xlabel('Time (min)')
        >>> ax0.set_ylabel('Tissue concentration (mM)')
        >>> ax0.legend()

        Plot WV results in the right panel:

        >>> ax1.set_title('Weakly vascularised model')
        >>> ax1.plot(t/60, Cwv*0, linestyle='-', linewidth=3.0, 
        >>>          color='darkred', label='Plasma')
        >>> ax1.plot(t/60, 1000*Cwv, linestyle='-', 
        >>>          linewidth=3.0, color='grey', label='Tissue')
        >>> ax1.set_xlabel('Time (min)')
        >>> ax1.set_ylabel('Tissue concentration (mM)')
        >>> ax1.legend()
        >>> plt.show()
    """

    if kinetics == 'U':
        return _conc_u(ca, t=t, dt=dt, **params)
    elif kinetics == 'FX':
        return _conc_fx(ca, t=t, dt=dt, **params)
    elif kinetics == 'NX':
        return _conc_nx(ca, t=t, dt=dt, **params)
    elif kinetics == 'NXP':
        return _conc_nxp(ca, t=t, dt=dt, **params)
    elif kinetics == 'WV':
        return _conc_wv(ca, t=t, dt=dt, **params)
    elif kinetics == 'HFU':
        return _conc_hfu(ca, t=t, dt=dt, sum=sum, **params)
    elif kinetics == 'HF':
        return _conc_hf(ca, t=t, dt=dt, sum=sum, **params)
    elif kinetics == '2CU':
        return _conc_2cu(ca, t=t, dt=dt, sum=sum, **params)
    elif kinetics == '2CX':
        return _conc_2cx(ca, t=t, dt=dt, sum=sum, **params)
    # elif model=='2CF':
    #     return _conc_2cf(ca, *params, t=t, dt=dt, sum=sum)
    else:
        raise ValueError(
            'Kinetic model ' + kinetics + ' is not currently implemented.')



def _conc_u(ca, t=None, dt=1.0, Fb=None):
    C = pk.conc_trap(Fb*ca, t=t, dt=dt)
    return C
    

def _conc_fx(ca, t=None, dt=1.0, 
             H=None, ve=None, Fb=None):
    if Fb == 0:
        ce = ca*0
    else:
        Fp = (1-H)*Fb
        ce = pk.flux_comp(ca/(1-H), ve/Fp, t=t, dt=dt)
    return ve*ce


def _conc_nx(ca, t=None, dt=1.0, 
             vb=None, Fb=None):
    if Fb == 0:
        Cp = ca*0
    else:
        Cp = pk.conc_comp(Fb*ca, vb/Fb, t=t, dt=dt)
    return Cp 

def _conc_nxp(ca, t=None, dt=1.0, 
             vb=None, Fb=None):
    if Fb == 0:
        Cp = ca*0
    else:
        Cp = pk.conc_plug(Fb*ca, vb/Fb, t=t, dt=dt)
    return Cp 


def _conc_wv(ca, t=None, dt=1.0, 
             H=None, vi=None, Ktrans=None):
    if Ktrans == 0:
        ci = ca*0
    else:
        ci = pk.flux_comp(ca/(1-H), vi/Ktrans, t=t, dt=dt)
    return vi*ci


def _conc_hfu(ca, t=None, dt=1.0, sum=True, 
              H=None, vb=None, PS=None):
    vp = vb*(1-H)
    cp = ca/(1-H)
    Ci = pk.conc_trap(PS*cp, t=t, dt=dt)
    if sum:
        return vp*cp + Ci
    else:
        return np.stack((vp*cp, Ci)) 


def _conc_hf(ca, t=None, dt=1.0, sum=True,
             H=None, vi=None, vb=None, PS=None):
    vp = vb*(1-H)
    ca = ca/(1-H)
    Cp = vp*ca
    if PS == 0:
        Ci = 0*ca
    else:
        Ci = pk.conc_comp(PS*ca, vi/PS, t=t, dt=dt)
    if sum:
        return Cp+Ci
    else:
        return np.stack((Cp, Ci))


def _conc_2cu(ca, t=None, dt=1.0, sum=True,
              H=None, vb=None, Fb=None, PS=None):
    vp = (1-H)*vb
    Fp = (1-H)*Fb
    if np.isinf(Fp):
        return _conc_hfu(ca, t=t, dt=dt, sum=sum, H=H, vb=vb, PS=PS)
    ca = ca/(1-H)
    if Fp+PS == 0:
        if sum:
            return np.zeros(len(ca))
        else:
            return np.zeros((2, len(ca)))
    Tp = vp/(Fp+PS)
    Cp = pk.conc_comp(Fp*ca, Tp, t=t, dt=dt)
    if vp == 0:
        Ktrans = PS*Fp/(PS+Fp)
        Ci = pk.conc_trap(Ktrans*ca, t=t, dt=dt)
    else:
        Ci = pk.conc_trap(PS*Cp/vp, t=t, dt=dt)
    if sum:
        return Cp+Ci
    else:
        return np.stack((Cp, Ci))


def _conc_2cx(ca, t=None, dt=1.0, sum=True,
              H=None, vi=None, vb=None, Fb=None, PS=None):

    vp = (1-H)*vb
    Fp = (1-H)*Fb
    if np.isinf(Fp):
        return _conc_hf(ca, t=t, dt=dt, sum=sum, 
                        H=H, vi=vi, vb=vb, PS=PS)

    ca = ca/(1-H)
    J = Fp*ca

    if Fp+PS == 0:
        Cp = np.zeros(len(ca))
        Ce = np.zeros(len(ca))
        if sum:
            return Cp+Ce
        else:
            return np.stack((Cp, Ce))

    Tp = vp/(Fp+PS)
    E = PS/(Fp+PS)

    if PS == 0:
        Cp = pk.conc_comp(Fp*ca, Tp, t=t, dt=dt)
        Ci = np.zeros(len(ca))
        if sum:
            return Cp+Ci
        else:
            return np.stack((Cp, Ci))

    Ti = vi/PS

    C = pk.conc_2cxm(J, [Tp, Ti], E, t=t, dt=dt)
    if sum:
        return np.sum(C, axis=0)
    else:
        return C
    

# def _conc_2cf(ca, t=None, dt=1.0, sum=True, 
#               vp=None, Fp=None, PS=None, Te=None):
#     if Fp+PS == 0:
#         if sum:
#             return np.zeros(len(ca))
#         else:
#             return np.zeros((2, len(ca)))
#     # Derive standard parameters
#     Tp = vp/(Fp+PS)
#     E = PS/(Fp+PS)
#     J = Fp*ca
#     T = [Tp, Te]
#     # Solve the system explicitly
#     t = utils.tarray(len(J), t=t, dt=dt)
#     C0 = pk.conc_comp(J, T[0], t)
#     if E == 0:
#         C1 = np.zeros(len(t))
#     elif T[0] == 0:
#         J10 = E*J
#         C1 = pk.conc_comp(J10, T[1], t)
#     else:
#         J10 = C0*E/T[0]
#         C1 = pk.conc_comp(J10, T[1], t)
#     if sum:
#         return C0+C1
#     else:
#         return np.stack((C0, C1))
    

# def _lconc_fx(ca, t=None, dt=1.0, Te=None):
#     # Te = ve/Fp
#     ce = pk.flux_comp(ca, Te, t=t, dt=dt)
#     return ce


# def _lconc_u(ca, t=None, dt=1.0, Tb=None):
#     # Tb = vp/Fp
#     if Tb==0:
#         msg = ('An uptake tissue with Tb=0 is not well-defined. \n'
#                 + 'Consider constraining the parameters.')
#         raise ValueError(msg)
#     cp = pk.conc_trap(ca, t=t, dt=dt)/Tb
#     return cp

    
# def _lconc_nx(ca, t=None, dt=1.0, Tb=None):
#     cp = pk.flux_comp(ca, Tb, t=t, dt=dt)
#     return cp

    
# def _lconc_wv(ca, t=None, dt=1.0, Ti=None):
#     # Note cp is non-zero and equal to (1-E)*ca
#     # But is not returned as it sits in a compartment without dimensions
#     # Ti = vi/Ktrans
#     ci = pk.flux_comp(ca, Ti, t=t, dt=dt)
#     return ci

        
# def _lconc_hfu(ca, t=None, dt=1.0, Ti=None):
#     # Ti=vi/PS
#     # up = vp / (vp + vi)
#     cp = ca
#     if Ti==0:
#         msg = 'An uptake tissue with Ti=0 is not well-defined. \n'
#         msg += 'Consider constraining the parameters.'
#         raise ValueError(msg)
#     ci = pk.conc_trap(cp, t=t, dt=dt)/Ti
#     return np.stack((cp, ci))

        
# def _lconc_hf(ca, t=None, dt=1.0, Ti=None):
#     # Ti = vi/PS
#     # up = vp/ve
#     cp = ca
#     ci = pk.flux_comp(cp, Ti, t=t, dt=dt)
#     return np.stack((cp, ci))


# def _lconc_2cu(ca, t=None, dt=1.0, 
#                Tp=None, E=None, Ti=None):
#     # Ti = vi/PS
#     cp = (1-E)*pk.flux_comp(ca, Tp, t=t, dt=dt)
#     ci = pk.conc_trap(cp, t=t, dt=dt)/Ti
#     return np.stack((cp, ci))

        
# def _lconc_2cx(ca, t=None, dt=1.0, Tp=None, Ti=None, E=None):
#     # c = C/Fp
#     # cp = C0/vp = c0*Fp/vp = c0 * (1-E)/Tp
#     # ci = C1/vi = c1*Fp/vi = c1 * (1-E)/E/Ti
#     c = pk.conc_2cxm(ca, [Tp, Ti], E, t=t, dt=dt)
#     cp = c[0,:] * (1-E) / Tp
#     ci = c[1,:] * (1-E) / E / Ti
#     return np.stack((cp, ci))


def flux_tissue(ca: np.ndarray, t=None, dt=1.0, kinetics='2CX', **params) -> np.ndarray:
    """Indicator flux out of a 2-site exchange tissue.

    Args:
        ca (array-like): concentration in the arterial input.
        t (array_like, optional): the time points of the input function *ca*. 
          If *t* is not provided, the time points are assumed to be uniformly 
          spaced with spacing *dt*. Defaults to None.
        dt (float, optional): spacing in seconds between time points for 
          uniformly spaced time points. This parameter is ignored if *t* is 
          provided. Defaults to 1.0.
        kinetics (str, optional): The kinetic model of the tissue (see below 
          for possible values). Defaults to '2CX'. 
        params (dict): free model parameters and their values (see below for 
          possible).

    Returns: 
        numpy.ndarray: outflux
          For a one-compartmental tissue, outflux out of the 
          compartment as a 1D array in units of mmol/sec/mL or M/sec. For a 
          multi=compartmental tissue, outflux out of each compartment, and at 
          each time point, as a 3D array with dimensions *(2,2,k)*, where *2* 
          is the number of compartments and *k* is the number of time points 
          in *J*. Encoding of the first two indices is the same as for *E*: 
          *J[j,i,:]* is the flux from compartment *i* to *j*, and *J[i,i,:]* 
          is the flux from *i* directly to the outside. The flux is returned in 
          units of mmol/sec/mL or M/sec.
    """
    if kinetics == 'U':
        return _flux_u(ca, **params)
    elif kinetics == 'NX':
        return _flux_nx(ca, t=t, dt=dt, **params)
    elif kinetics == 'FX':
        return _flux_fx(ca, t=t, dt=dt, **params)
    elif kinetics == 'WV':
        return _flux_wv(ca, t=t, dt=dt, **params)
    elif kinetics == 'HFU':
        return _flux_hfu(ca, **params)
    elif kinetics == 'HF':
        return _flux_hf(ca, t=t, dt=dt, **params)
    elif kinetics == '2CU':
        return _flux_2cu(ca, t=t, dt=dt, **params)
    elif kinetics == '2CX':
        return _flux_2cx(ca, t=t, dt=dt, **params)
    # elif model=='2CF':
    #     return _flux_2cf(ca, t=t, dt=dt, **params)
    else:
        raise ValueError('Kinetic model ' + kinetics +
                         ' is not currently implemented.')


def _flux_u(ca, Fb=None):
    return pk.flux(Fb*ca, model='trap')


def _flux_nx(ca, t=None, dt=1.0, vb=None, Fb=None):
    if Fb == 0:
        return np.zeros(len(ca))
    return pk.flux(Fb*ca, vb/Fb, t=t, dt=dt, model='comp')

def _flux_fx(ca, t=None, dt=1.0, H=None, ve=None, Fb=None):
    if Fb == 0:
        return np.zeros(len(ca))
    Fp = Fb*(1-H)
    return pk.flux(Fb*ca, ve/Fp, t=t, dt=dt, model='comp')

def _flux_wv(ca, t=None, dt=1.0, H=None, vi=None, Ktrans=None):
    ca = ca/(1-H)
    J = np.zeros(((2, 2, len(ca))))
    J[0, 0, :] = np.nan
    J[1, 0, :] = Ktrans*ca
    if Ktrans != 0:
        J[0, 1, :] = pk.flux(Ktrans*ca, vi/Ktrans, t=t, dt=dt, model='comp')
    return J

def _flux_hfu(ca, H=None, PS=None):
    J = np.zeros(((2, 2, len(ca))))
    J[0, 0, :] = np.nan
    J[1, 0, :] = PS*ca/(1-H)
    return J


def _flux_hf(ca, t=None, dt=1.0, H=None, vi=None, PS=None):
    ca = ca/(1-H)
    J = np.zeros(((2, 2, len(ca))))
    J[0, 0, :] = np.inf
    J[1, 0, :] = PS*ca
    if PS == 0:
        J[0, 1, :] = 0*ca
    else:
        J[0, 1, :] = pk.flux(PS*ca, vi/PS, t=t, dt=dt, model='comp')
    return J


def _flux_2cu(ca, t=None, dt=1.0, H=None, vb=None, Fb=None, PS=None):
    C = _conc_2cu(ca, t=t, dt=dt, sum=False, H=H, vb=vb, Fb=Fb, PS=PS)
    ca = ca/(1-H)
    Fp = Fb*(1-H)
    J = np.zeros(((2, 2, len(ca))))
    if vb == 0:
        if Fp+PS != 0:
            Ktrans = Fp*PS/(Fp+PS)
            J[0, 0, :] = Fp*ca
            J[1, 0, :] = Ktrans*ca
    else:
        J[0, 0, :] = Fp*C[0, :]/vb
        J[1, 0, :] = PS*C[0, :]/vb
    return J


def _flux_2cx(ca, t=None, dt=1.0, H=None, vb=None, vi=None, Fb=None, PS=None):

    if np.isinf(Fb):
        return _flux_hf(ca, t=t, dt=dt, H=H, vi=vi, PS=PS)

    if Fb == 0:
        return np.zeros((2, 2, len(ca)))

    Fp = Fb*(1-H)

    if PS == 0:
        Jp = _flux_nx(ca, t=t, dt=dt, vb=vb, Fb=Fb)
        J = np.zeros((2, 2, len(ca)))
        J[0, 0, :] = Jp
        return J
    C = _conc_2cx(ca, t=t, dt=dt, sum=False, H=H, vb=vb, vi=vi, Fb=Fb, PS=PS)
    # Derive standard parameters
    vp = vb*(1-H)
    Tp = vp/(Fp+PS)
    Te = vi/PS
    E = PS/(Fp+PS)
    # Build the system matrix K
    T = [Tp, Te]
    E = [
        [1-E, 1],
        [E,   0],
    ]
    return pk._J_ncomp(C, T, E)


# def _flux_2cf(ca, t=None, dt=1.0, vp=None, Fp=None, PS=None, Te=None):
#     if Fp+PS == 0:
#         return np.zeros((2, 2, len(ca)))
#     # Derive standard parameters
#     Tp = vp/(Fp+PS)
#     E = PS/(Fp+PS)
#     J = Fp*ca
#     T = [Tp, Te]
#     # Solve the system explicitly
#     t = utils.tarray(len(J), t=t, dt=dt)
#     Jo = np.zeros((2, 2, len(t)))
#     J0 = pk.flux(J, T[0], t=t, model='comp')
#     J10 = E*J0
#     Jo[1, 0, :] = J10
#     Jo[1, 1, :] = pk.flux(J10, T[1], t=t, model='comp')
#     Jo[0, 0, :] = (1-E)*J0
#     return Jo



import os

import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm

from dcmri import rel, sig, pk_inv, pk_aorta



PARAMS = {
    'dt': {
        'init': 1.0,
        'name': 'Time step',
        'unit': 'sec',
    },
    'ca': {
        'init': None,
        'name': 'Arterial Input concentration',
        'unit': 'mL/sec/cm3',
    },
    'irf': {
        'init': None, 
        'name': 'Impulse response function',
        'unit': 'mL/sec/cm3',
    },
    'r1': {
        'init': 5000.0,
        'name': 'Contrast agent relaxivity',
        'unit': 'Hz/M',
    },
    'R10a': {
        'init': 0.7,
        'name': 'Arterial precontrast R1',
        'unit': 'Hz',
    },
    'S0a': {
        'init': 1.0,
        'name': 'Arterial signal scaling factor',
        'unit': 'a.u.',
    },
    'B1corr_a': {
        'init': 1,
        'name': 'Arterial B1-correction factor',
        'unit': '',
    },
    'B1corr': {
        'name': 'Tissue B1-correction factor',
        'unit': '',
    },
    'FA': {
        'init': 15,
        'name': 'Flip angle',
        'unit': 'deg',
    },
    'TR': {
        'init': 0.005,
        'name': 'Repetition time',
        'unit': 'sec',
    },
    'TC': {
        'init': 0.2,
        'name': 'Time to k-space center',
        'unit': 'sec',
    },
    'TS': {
        'init': 0,
        'name': 'Sampling time',
        'unit': 'sec',
    },
    'R10': {
        'name': 'Tissue precontrast R1',
        'unit': 'Hz',
    },
    'S0': {
        'name': 'Signal scaling factor',
        'unit': 'a.u.',
    },
    'H': {
        'init': 0.45,
        'name': 'Tissue Hematocrit',
        'unit': '',
    },
    'Fb': {
        'name': 'Parenchymal blood flow',
        'unit': 'mL/sec/cm3',
    },
    've': {
        'name': 'Extracellular volume',
        'unit': 'mL/cm3',
    },
    'Te': {
        'name': 'Extracellular mean transit time',
        'unit': 'sec',
    },
}





class TissueLS():
    """Array of linear and stationary tissues with a single inlet.

    These are generic model-free tissue types. Their response to 
    an indicator injection is proportional to the dose (linear) and 
    independent of the time of injection (stationary).

    Args:
        shape (array-like, required): shape of the tissue array (spatial dimensions only). 
          Any number of dimensions is allowed.
        aif (array-like, required): Signal-time curve in the blood of the
          feeding artery. 
        dt (float, optional): Time interval between values of the arterial
          input function. Defaults to 1.0.
        sequence (str, optional): imaging sequence. Possible values 
          are 'SS', 'SR' and 'lin' (linear). Defaults to 'SS'.
        params (dict, optional): values for the parameters of the tissue,
          specified as keyword parameters. Defaults are used for any that are
          not provided. 

    See Also:
        `TissueLS`, `TissueArray`

    Example:

        Fit a linear and stationary model to the synthetic test data:

    .. plot::
        :include-source:
        :context: close-figs

        >>> import numpy as np
        >>> import dcmri as dc

        Generate synthetic test data:

        >>> time, aif, roi, gt = dc.fake_tissue()

        The correct ground truth for ve in model-free analysis is the 
        extracellular part of the distribution space:

        >>> gt['ve'] = gt['vp'] + gt['vi'] if gt['PS'] > 0 else gt['vp']

        Build a tissue and set the constants to match the
        experimental conditions of the synthetic test data. 

        >>> tissue = dc.TissueLS(
        ...     dt = time[1],
        ...     sequence = 'SS',
        ...     r1 = dc.relaxivity(3, 'blood','gadodiamide'),
        ...     TR = 0.005,
        ...     FA = 15,
        ...     R10a = 1/dc.T1(3.0,'blood'),
        ...     R10 = 1/dc.T1(3.0,'muscle'),
        ... )

        Train the tissue on the data. Since have noise-free synthetic 
        data we use a lower tolerance than the default, which is optimized 
        for noisy data:

        >>> tissue.train(roi, aif, n0=10, tol=0.01)

        Plot the reconstructed signals along with the concentrations 
        and the impulse response function.

        >>> tissue.plot(roi)

    """

    def __init__(self, sequence='SS', **kwargs):

        # Configuration
        if sequence not in ['SS', 'SR', 'lin']:
            raise ValueError(
                f"Sequence {sequence} is not recognized. "
                f"Current options are 'SS', 'SR', 'lin'."
            )
        self.sequence = sequence
        self.pars = {}
        
        # Initialize scalar parameters
        params = ['dt', 'H', 'R10a', 'S0a', 'r1']
        if self.sequence == 'SR':
            params += ['TC']
        elif self.sequence == 'SS':
            params += ['TR', 'B1corr_a', 'FA']
        elif self.sequence == 'lin':
            params += []
        for par in params:
            self.pars[par] = PARAMS[par]['init']

        # Initialize array parameters
        nt = 120
        time = self.pars['dt'] * np.arange(nt)
        Kb = 1 / 5.0
        self.pars['ca'] = pk_aorta.aif_tristan(time)
        self.pars['irf'] = 0.01 * np.exp(-Kb * time)
        self.pars['S0'] = 1
        self.pars['R10'] = 1
        if self.sequence == 'SS':
            self.pars['B1corr'] = 1

        # Override parameter defaults
        for par in kwargs:
            self.pars[par] = kwargs[par]


    def predict_aif(self):
        """Predict the signal at specific time points

        Returns:
            np.ndarray: Array of predicted signals for each time point.
        """
        # Predict arterial signal
        R1a = rel.relax(self.pars['ca'], self.pars['R10a'], self.pars['r1'])
        if self.sequence == 'SS':
            Sa = sig.signal_ss(self.pars['S0a'], R1a, self.pars['TR'], self.pars['B1corr']*self.pars['FA'])
        elif self.sequence == 'SR':
            Sa = sig.signal_src(self.pars['S0a'], R1a, self.pars['TC'])
        elif self.sequence == 'lin':
            Sa = sig.signal_lin(self.pars['S0a'], R1a) 
        return Sa
 
    def predict_conc(self):
        """Return the tissue concentration

        Returns:
            np.ndarray: Concentration in M
        """
        ca_mat = self.pars['dt'] * pk_inv.convmat(self.pars['ca'])
        irf_mat = self.pars['irf']
        conc = ca_mat @ irf_mat.T
        return conc.T

    def predict(self):
        """Predict the signal at specific time points

        Returns:
            np.ndarray: Array of predicted signals for each time point.
        """ 
        conc = self.predict_conc()
        R1 = rel.relax(conc, self.pars['R10'], self.pars['r1'])
        S0 = self.pars['S0']
        if self.sequence == 'SS':
            signal = sig.signal_ss(S0, R1, self.pars['TR'], self.pars['B1corr']*self.pars['FA'])
        elif self.sequence == 'SR':
            signal = sig.signal_src(S0, R1, self.pars['TC'])
        elif self.sequence == 'lin':
            signal = sig.signal_lin(S0, R1) 
        return signal

    

    def train(self, signal, signal_aif, n0=1, tol=0.1, init_s0=True):
        """Train the free parameters

        Args:
            signal (array-like): Array with measured signals.
            tol: cut-off value for the singular values in the 
                computation of the matrix pseudo-inverse.

        Returns:
            self
        """ 

        # Fit baselines if needed
        if init_s0:
            if self.sequence == 'SR':
                scla = sig.signal_src(1, self.pars['R10a'],  self.pars['TC'])
                scl = sig.signal_src(1, self.pars['R10'],  self.pars['TC'])
            elif self.sequence == 'SS':
                scla = sig.signal_ss(1, self.pars['R10a'],  self.pars['TR'], self.pars['B1corr_a'] * self.pars['FA'])
                scl = sig.signal_ss(1, self.pars['R10'], self.pars['TR'], self.pars['B1corr'] * self.pars['FA'])
            elif self.sequence == 'lin':
                scla = sig.signal_lin(1, self.pars['R10a'])
                scl = sig.signal_lin(1, self.pars['R10'])
            self.pars['S0a'] = np.mean(signal_aif[:n0]) / scla if scla > 0 else 0
            self.pars['S0'] = np.mean(signal[...,:n0]) / scl if scl > 0 else 0

        # Derive concentrations
        T10a = 1/self.pars['R10a'] if self.pars['R10a'] > 0 else 0
        T10 = 1/self.pars['R10'] if self.pars['R10'] > 0 else 0

        if self.sequence == 'SR':
            self.pars['ca'] = sig.conc_src(
                signal_aif, self.pars['TC'], T10a, 
                self.pars['r1'], S0=self.pars['S0a'])
            conc = sig.conc_src(
                signal, self.pars['TC'], T10, 
                self.pars['r1'], S0=self.pars['S0'])
            
        elif self.sequence == 'SS':
            self.pars['ca'] = sig.conc_ss(
                signal_aif, self.pars['TR'], self.pars['B1corr_a'] * self.pars['FA'],
                T10a, self.pars['r1'], S0=self.pars['S0a'])
            conc = sig.conc_ss(
                signal, self.pars['TR'], self.pars['B1corr_a'] * self.pars['FA'],
                T10, self.pars['r1'], S0=self.pars['S0'])
            
        elif self.sequence == 'lin':
            self.pars['ca'] = sig.conc_lin(
                signal_aif, T10a, 
                self.pars['r1'], S0=self.pars['S0a'])
            conc = sig.conc_lin(
                signal, T10, self.pars['r1'], S0=self.pars['S0'])

        conc[np.isnan(conc)] = 0
        irf_mat = pk_inv.deconv(conc, self.pars['ca'], self.pars['dt'], tol=tol)
        self.pars['irf'] = irf_mat

        return self
    
    
    def params(self, *args):
        """Export the tissue parameters

        Args:
            args (tuple): parameters to get. If no arguments are 
                provided, all available parameters are returned.    

        Returns:
            dict: Dictionary with tissue parameters.
        """
        amax = np.max(self.pars['irf'])
        auc = np.sum(self.pars['irf']) * self.pars['dt']
        params = {
            'IRF': self.pars['irf'],
            'Fb': amax,
            'Te': auc / amax if amax > 0 else 0,
            've': auc * (1-self.pars['H']),
            'S0': self.pars['S0'],
        }
        if args == ():
            return params
        elif len(args) == 1:
            return params[args[0]]
        else:
            return {p: params[p] for p in args}


    def print_params(self, round_to=None):
        """Print the model parameters

        Args:
            round_to (int, optional): Round to how many digits. If this is
              not provided, the values are not rounded. Defaults to None.
        """

        print('')
        print('----------------------------------------------------')
        print('Derived parameters for linear and stationary tissues')
        print('----------------------------------------------------')
        print('')  

        p = self.params()

        par = p['Fb']
        par = round(par, round_to) if round_to is not None else par
        print(f"Parenchymal blood flow: {par} mL/sec/cm3")  

        par = p['Te']
        par = round(par, round_to) if round_to is not None else par 
        print(f"Extracellular transit time: {par} sec") 

        par = p['ve']
        par = round(par, round_to) if round_to is not None else par
        print(f"Extracellular volume fraction: {par} mL/cm3")



    def plot(self, signal=None, round_to=None, fname=None, show=True):
        """Plot the model fit against data

        Args:
            signal (array-like, optional): Array with measured signals.
            round_to (int, optional): Rounding for the model parameters.
            fname (path, optional): Filepath to save the image. If no value is
              provided, the image is not saved. Defaults to None.
            show (bool, optional): If True, the plot is shown. Defaults to
              True.
        """
        
        fig, ax = plt.subplots(1, 4, figsize=(20, 5))
    
        t = self.pars['dt'] * np.arange(len(self.pars['ca']))

        # Plot predicted signals and measured signals
        ax[0].set_title('MRI signals')
        ax[0].set(ylabel='MRI signal (a.u.)', xlabel='Time (min)')
        ax[0].plot(
            t / 60, self.predict(), linestyle='-', 
            linewidth=3.0, color='cornflowerblue', 
            label='Tissue (predicted)',
        )
        if signal is not None:
            ax[0].plot(
                t / 60, signal, marker='x', linestyle='None',
                color='darkblue', label='Tissue (measured)',
            )
        ax[0].legend()

        # Plot predicted concentrations and measured concentrations
        ax[1].set_title('Tissue concentrations')
        ax[1].set(ylabel='Concentration (mM)', xlabel='Time (min)')
        ax[1].plot(
            t / 60, 1000 * self.pars['ca'], linestyle='-', linewidth=5.0,
            color='lightcoral', label='Arterial blood',
        )
        ax[1].plot(
            t / 60, 1000 * self.predict_conc(), linestyle='-', linewidth=3.0, 
            color='cornflowerblue', label='Tissue (predicted)', 
        )
        ax[1].legend()

        # Plot impulse response
        ax[2].set_title('Impulse response function')
        ax[2].set(ylabel='IRF (mL/sec/cm3)', xlabel='Time (min)')
        ax[2].plot(
            t / 60, self.pars['irf'], linestyle='-', linewidth=3.0,
            color='cornflowerblue', label='IRF', 
        )
        ax[2].legend()

        # Plot text
        pars = self.params()
        msg = []
        for par in ['Fb', 've', 'Te', 'S0']:
            value = pars[par]
            if round_to is not None:
                value = round(value, round_to)
            msg.append(f"{PARAMS[par]['name']} ({par}): {value} {PARAMS[par]['unit']} \n")
        
        msg = "\n".join(msg)
        ax[3].set_title('Free parameters')
        ax[3].axis("off")  # hide axes
        ax[3].text(0, 0.9, msg, fontsize=10, transform=ax[3].transAxes, ha="left", va="top")

        # Show and/or save plot
        if fname is not None:
            plt.savefig(fname=fname)
        if show:
            plt.show()
        else:
            plt.close()

        


import numpy as np

import dcmri as dc



def test_convmat():

    # Uniform time grid with increasing precision: compare against analytical convolution.

    Tf = 20
    Th = 30
    tmax = 30

    prec = [1e-1, 1e-3, 1e-5]
    for i, dt in enumerate([10,1,0.1]):
        t = np.arange(0,tmax,dt)
        f = np.exp(-t/Tf)/Tf
        h = np.exp(-t/Th)/Th
        mat = dc.convmat(f)
        #g = dt*np.matmul(mat.T, h)
        g = dt * (mat @ h)
        g0 = (Tf*f-Th*h)/(Tf-Th)
        assert np.linalg.norm(g-g0)/np.linalg.norm(g0) < prec[i]
        mat = dc.convmat(f, order=1)
        g1 = dt * (mat @ h)
        assert np.linalg.norm(g-g0) < np.linalg.norm(g1-g0)


def test_invconvmat():

    Tf = 20
    tmax = 30
    dt = 0.1
    order = 1
    t = np.arange(0,tmax,dt)
    f = np.exp(-t/Tf)/Tf
    mat = dc.convmat(f, order=order)
    matinv = dc.invconvmat(f, order=order, tol=1e-12)
    id = mat @ matinv
    idexact = np.eye(len(t))
    assert np.linalg.norm(id-idexact)/np.linalg.norm(idexact) < 1e-9


def test_deconv():

    # Uniform time grid with increasing precision: compare against analytical convolution.

    Tf = 20
    Tg = 30
    tmax = 100
    dt = 0.1
    t = np.arange(0,tmax,dt)
    f = np.exp(-t/Tf)/Tf
    g = np.exp(-t/Tg)/Tg
    h = (dc.convmat(g) @ f) * dt
    frec = dc.deconv(h, g, dt)
    assert np.linalg.norm(f-frec)/np.linalg.norm(f) < 0.1

    F = np.tile(f[:, None], (1, 3))
    H = np.tile(h[:, None], (1, 3))
    Frec = dc.deconv(H, g, dt)
    assert np.linalg.norm(F-Frec)/np.linalg.norm(F) < 0.1


if __name__ == "__main__":

    print('Testing pk_inv..')

    test_convmat()
    test_invconvmat()
    test_deconv()

    print('All pk_inv tests passed!!')
